/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/


/*File arpissues.h.

This file contains functions related to ARP packets*/

void ARPreply (unsigned long, char *, unsigned long, char *);


typedef struct arhdr{
  
        unsigned short  ar_hrd;		/* format of hardware address	*/
	unsigned short	ar_pro;		/* format of protocol address	*/
	unsigned char	ar_hln;		/* length of hardware address	*/
	unsigned char	ar_pln;		/* length of protocol address	*/
	unsigned short	ar_op;		/* ARP opcode (command)		*/
        unsigned char ar_sha[ETH_ALEN]; /* sender hardware address	*/
	unsigned char ar_sip[4];	/* sender IP address		*/
	unsigned char ar_tha[ETH_ALEN];	/* target hardware address	*/
	unsigned char ar_tip[4];       	/* target IP address		*/

      } Arphdr;

/*this struct defines all the fields for ARP packets*/

/*********************************************************************/

int requestARP (unsigned long Address, char *hwdata)
{

/*Given an IP address (specified in Address) it stores in hwdata the MAC address 
of that IP address (if the IP address is in the ARP cache) and it returns 1.
Otherwise it returns 0*/

  int ioctlcall, value;
  struct sockaddr_in *sock;
  struct arpreq request;

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

  bzero ((char *)&request, sizeof (struct arpreq));

  sock = (struct sockaddr_in *)&request.arp_pa;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;

  sock = (struct sockaddr_in *)&request.arp_ha;
  sock->sin_family = ARPHRD_ETHER;

  sock = (struct sockaddr_in *)&request.arp_netmask;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = inet_addr ("255.255.255.255");

  if (value =ioctl(ioctlcall, SIOCGARP, (char *)&request) < 0)
     DEBUG (3) 
       {
         perror ("Get ARP error issued");
         close (ioctlcall);
	 return 0;
       }
  close (ioctlcall);
  bcopy ((char *)request.arp_ha.sa_data, hwdata, 6);
  return 1;
}

/********************************************************************/

void createARPPerm (unsigned long Address)
{

/*It creates an ARP entry as permanent in the ARP cache given . It checks first if the
entry is already in the ARP cache*/

  int ioctlcall;
  struct sockaddr_in *sock;
  struct arpreq request;

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
 
  bzero ((char *)&request, sizeof (struct arpreq));
  sock = (struct sockaddr_in *) &request.arp_pa;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;

  sock = (struct sockaddr_in *)&request.arp_ha;
  sock->sin_family = ARPHRD_ETHER;
  if (!requestARP (Address, request.arp_ha.sa_data))
   {
     DEBUG (1) perror ("I cannot find ARP entry");
     close (ioctlcall);
     return;
   }
  request.arp_flags =  (ATF_PERM | ATF_COM);
  if (ioctl(ioctlcall, SIOCSARP, (char *)&request) < 0)
      DEBUG (1) perror ("ARP has not been done");
  close (ioctlcall);
}

/********************************************************************/

int hwaddread(char s[], char *hwadr)  
{
   int a, b, c, d, e, f;
   if (sscanf(hwadr,"%x:%x:%x:%x:%x:%x", &a, &b, &c, &d, &e, &f) != 6)
     return(-1);
   else  
   {
      s[0] = (unsigned char) a;
      s[1] = (unsigned char) b;
      s[2] = (unsigned char) c;
      s[3] = (unsigned char) d;
      s[4] = (unsigned char) e;
      s[5] = (unsigned char) f;
     return(0);
   }
}

/*********************************************************************/
 
void GratuitousARP (unsigned long ipaddr,char *hwaddr)
{

/*This function is intended to send a gratuitous ARP packet with IP address
ipaddr and hardware address hwaddr to the interface defined in interface*/

  char frame[60];
  struct ethhdr *ehdr;
  Arphdr *ahdr;
  int sockid;
  struct sockaddr sa;
  
  if((sockid = socket(AF_INET,SOCK_PACKET,htons(ETH_P_802_3)))<0)
    {
      perror("Socket call failed:");
      exit(-1);
    }

  bzero((void*)frame,sizeof(frame));

/* Ethernet header */

  ehdr = (struct ethhdr*)frame;
  hwaddread(ehdr->h_dest,"ff:ff:ff:ff:ff:ff");

  bcopy(hwaddr,ehdr->h_source,6);
  ehdr->h_proto = htons(ETH_P_ARP);

  ahdr = (Arphdr*)(frame + sizeof(struct ethhdr));

/* Arp header */

  ahdr->ar_hrd = htons(ARPHRD_ETHER);
  ahdr->ar_pro = htons(ETH_P_IP);
  ahdr->ar_hln = 6;
  ahdr->ar_pln = 4;
  ahdr->ar_op  = htons(ARPOP_REQUEST);

  bcopy(hwaddr,ahdr->ar_sha,6);
  bcopy(&ipaddr,ahdr->ar_sip,4);
  bzero(ahdr->ar_tha,6);
  bcopy(&ipaddr,ahdr->ar_tip,4);
  
   
  sa.sa_family = AF_INET;
  strcpy(sa.sa_data, interface);
  printf ("Interface for the gratuitous ARP %s\n", interface);
  
  if(sendto(sockid, frame, sizeof(frame), 0, &sa, sizeof(sa))<0)
    {
      perror("Sendto failed");
      exit(-1);
    }

  close (sockid);
  
}

/*********************************************************************/
#ifdef ARP_REPLIES
void ARP_send_background ()
{

 ARPreply (mynode.FA.COA, mynode.FA.hwAddress, mynode.Address,
		mynode.MAC);

 mynode.ARPRepTimeout = addTimer(300, ARP_send_background, NULL, NULL);

}
#endif
/*********************************************************************/

void gratARPbackground (void *ipaddress, void *hwaddr)
{

/*It sends gratuitous ARP. It's used in the background*/

   unsigned long *ip;
   char *hw;
   
   ip = (unsigned long *)ipaddress;
   hw = (char *)hwaddr;
   
   GratuitousARP (*ip, hw);
}

/*********************************************************************/

void ARPreply (unsigned long COA, char *COAMAC, unsigned long myaddress,
		char *myMACaddr)
{  

/*It sends ARP replies to the FA (it's used in the mobile node) in order
to give it mobile node's MAC address*/

  char frame[60];
  struct ethhdr *ehdr;
  Arphdr *ahdr;
  int sockid;
  struct sockaddr sa;
  
  if((sockid = socket(AF_INET,SOCK_PACKET,htons(ETH_P_802_3)))<0)
    {
      perror("Socket call failed:");
      exit(-1);
    }

  bzero((void*)frame,sizeof(frame));
DEBUG (5)
{
  printf ("Sending an ARP reply\n");
  printf ("MAC of the FA: %x:%x:%x:%x:%x:%x\n", COAMAC[0], COAMAC[1], COAMAC[2],
	   COAMAC[3], COAMAC[4], COAMAC[5]);
  printf ("IP care of address %s ", inet_ntoa(COA));
  printf ("IP address of the source %s", inet_ntoa (myaddress));
  printf ("MAC address of the source %x:%x:%x:%x:%x:%x\n",myMACaddr[0], myMACaddr[1],
		myMACaddr[2], myMACaddr[3], myMACaddr[4], myMACaddr[5]);
}
/* Ethernet header */

  ehdr = (struct ethhdr*)frame;
  bcopy (COAMAC, ehdr->h_dest, 6);
  bcopy(myMACaddr,ehdr->h_source,6);
  ehdr->h_proto = htons(ETH_P_ARP);

  ahdr = (Arphdr*)(frame + sizeof(struct ethhdr));

/* Arp header */

  ahdr->ar_hrd = htons(ARPHRD_ETHER);
  ahdr->ar_pro = htons(ETH_P_IP);
  ahdr->ar_hln = 6;
  ahdr->ar_pln = 4;
  ahdr->ar_op  = htons(ARPOP_REPLY);

  bcopy(myMACaddr,ahdr->ar_sha,6);
  bcopy(&myaddress,ahdr->ar_sip,4);
  bcopy(COAMAC,ahdr->ar_tha,6);
  bcopy(&COA,ahdr->ar_tip,4);
  
   
  sa.sa_family = AF_INET;
  strcpy(sa.sa_data, interface);
  
  if(sendto(sockid, frame, sizeof(frame), 0, &sa, sizeof(sa))<0)
    {
      perror("Sendto failed:");
      exit(-1);
    }

  close (sockid);
  
}
