/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/


/*This program is used by the agents.
It is in charge of sending agent advertisements and reply when an
agent solicitaion is issued by mobile nodes.

Last modification 3/10/96*/

#include "includes_ag.h"

#define LIFETIME 60
#define LENGTHAGENTADVERT 16
#define MULTICAST "224.0.0.1"
#define DATAFILE "configuration"

#define HFLAG 0x2000
#define FFLAG 0x1000
#define RFLAG 0x8000


struct FAvalues ForeignA;
struct HAvalues HomeA;
struct authentication *authen;
struct MNode mynodes[10];

int SockAdverts;
char packet[1024];
char interface[5];

#include "parameters.h"

void routeEntryDev (unsigned long Address,char *device, unsigned long Mask)
{

  /*This procedure creates a route entry with destination address specified 
    in Address and it it sent to the interface specified in device*/

  struct sockaddr_in *sock, *sockmask;
  struct rtentry request;
  int ioctlcall;
  int count;

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
  bzero ((char *)&request, sizeof (struct rtentry));
  sock = (struct sockaddr_in *) &request.rt_dst;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;
  
  sockmask = (struct sockaddr_in *) &request.rt_genmask;
  sockmask->sin_family = AF_INET;
  sockmask->sin_addr.s_addr = Mask;

  request.rt_flags = (RTF_UP | RTF_HOST);
  request.rt_dev = device;
  if (ioctl(ioctlcall, SIOCADDRT, (char *)&request) == -1)
    perror ("Route entry can not be added");
  close (ioctlcall);
}

/********************************************************************/


void initsockets()
{

  /*This procedure creates a socket and it specifies the nature of the socket
    (multicast and able to issue broadcast packets*/

   int value, enable = 1;
   struct protoent *proto;
   struct ip_mreq multicast;
   
   if ((proto = getprotobyname("icmp")) == NULL) 
   {
                fprintf(stderr, "unknown protocol icmp\n");
                exit(1);
   } 

   SockAdverts = socket(AF_INET, SOCK_RAW, proto->p_proto);

   if (SockAdverts  < 0)  
   	{
	  fprintf(stderr, "initsockets(): could not create socket.\n");
	  exit(-1);
	}
              
   /*Multicast*/
   bzero ((char *)&multicast, sizeof (struct ip_mreq));
   multicast.imr_multiaddr.s_addr = inet_addr (MULTICAST);
   multicast.imr_interface.s_addr = ForeignA.COA;

   if ((value = setsockopt (SockAdverts, IPPROTO_IP, IP_ADD_MEMBERSHIP, &multicast,
                           sizeof (struct ip_mreq))) < 0)
    {
       printf ("Error: It cannot join the group!\n");
       exit (-1);
    }


   if (setsockopt (SockAdverts, SOL_SOCKET, SO_BROADCAST, (char *) &enable, sizeof (int)) < 0)
    {
      perror ("Broadcast error");
      exit (-1);
    }
}

/********************************************************************/

void sendAdvert()
{
  /*This procedure sends agent advertisements acording to the specifications
    of mobile IP*/


char buffer[1024];
struct icmphdr  *icp;
int i, length; 
struct advert *Ad;
struct sockaddr whereto;
struct sockaddr_in *to;

bzero ((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;

to->sin_addr.s_addr = htonl(INADDR_BROADCAST);

bzero (buffer,1024);
icp = (struct icmphdr *)buffer;
icp->type = 9;
icp->code = 0; /*It could be 16*/  /*Check it*/
icp->checksum = 0;      /* compute checksum below */
Ad = (struct advert *)(buffer+4);
Ad->num_addr = 0; /*Number of router addresses*/
Ad->addr_entry = 2; /*Size of each router entry (32 bit words)*/
Ad->Lifetime = htons(4); /*htons(ForeignA.Lifetime);*/ 
Ad->type=16;
Ad->length=10;
Ad->seq_num = htons(ForeignA.seq_num); 
Ad->Life_advert = htons(ForeignA.Lifetime); /*Max time between registrations?*/
Ad->flags = htons(ForeignA.flags); 
Ad->COA=ForeignA.COA;

length = 4 + sizeof (struct advert);

/* compute ICMP checksum here */

icp->checksum = in_cksum((u_short *)buffer, length);

i = sendto(SockAdverts, (char *)buffer, length, 0, &whereto, sizeof (struct sockaddr));

if (i < 0 || i != length) perror("sendto error");

DEBUG(4)
 {
   printf ("Agent Advert. sent! COA %s flags 0x%04lx seq num %d",inet_ntoa(ForeignA.COA),
           ForeignA.flags, ForeignA.seq_num);
   printf ("\nLifetime %d\n",ForeignA.Lifetime); 
 }
DEBUG (0) printf (".\n");
/*Draft section 2.3.2. It generates the next seq_num */
if (ForeignA.seq_num == 0xffff) ForeignA.seq_num = 256;
else ForeignA.seq_num++;
}

/********************************************************************/

void initall()
{

  /*This procedure initialites all the values*/

unsigned char temp[20];
int i;

loadparameters(DATAFILE);
initsockets();
ForeignA.seq_num=0;
ForeignA.Lifetime=LIFETIME; 
ForeignA.TimeBetweenAdverts = /*ForeignA.Lifetime/6 - 1 */ 2;
ForeignA.flags = HFLAG | FFLAG | RFLAG;
routeEntryDev((unsigned long)inet_addr("255.255.255.0"),interface,
	       (unsigned long)inet_addr("255.255.255.0"));
}

/********************************************************************/

void sendingAdverts()
{
	sendAdvert();
	signal(SIGALRM, (void *)sendingAdverts);
   	alarm(ALARMTIME);
}

/********************************************************************/

void checkAgentSol (char *packet, int length)
{

  /* When an agent solicitation is received this procedure checks it is 
     valid an it sends an agent advertisement*/

   struct icmphdr *icmp;
   struct iphdr *ip;
   int lengthIp;

   ip = (struct iphdr *)packet;
   lengthIp = ip->ihl <<2;
   icmp = (struct icmphdr *)(packet + lengthIp);
   if (length < lengthIp +8 || ip->protocol != 1 ||icmp->type !=10
	|| icmp->code != 0)  return;
   DEBUG (0) printf ("I've got an ICMP Agent Solicitation\n");
   sendAdvert(); 
}

/********************************************************************/

main(int argc, char *argv[])
{

   fd_set fdvector;
   struct timeval time;
   int selcode;
   int packlen = 1024;

   if (argc == 2) debugLevel = atoi(argv[1]);

   initall();
   signal(SIGALRM, (void *)sendingAdverts);
   alarm(ALARMTIME);
   for (;;)
    {
	FD_ZERO (&fdvector);
	FD_SET (SockAdverts, &fdvector);
	time.tv_sec = 0; time.tv_usec = 0;
	while ((selcode = select (FD_SETSIZE, &fdvector, NULL, NULL, &time)) == -1)
	  if (( errno == EINTR) || (errno == ERESTART));
	if (FD_ISSET(SockAdverts, &fdvector))
	  {
		struct sockaddr_in from;
		int length, fromlen;

		fromlen = sizeof (from);

		if ((length = recvfrom(SockAdverts, (char *)packet, packlen, 0, 
			(struct sockaddr *)&from, &fromlen)) <0)
		  {
			if (errno == EINTR) continue;
			DEBUG (0) perror ("rcvfrom error");
		  }
		DEBUG (3) printf ("I've got an ICMP packet\n");
		checkAgentSol (packet, length);
	  }
    }
}	


