/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/

/*This module, once is compiled, is the deamon which is running in the
agents*/

#include "includes_ag.h"

#define AGENT_LISTEN_PORT 434   /*Predefined port for Mobile IP*/
#define LIFETIME 60            /*This value can be changed*/
#define MAX_PACKET (65536 - 60 - 8)/* max packet size */
#define LENGTHAGENTADVERT 16
#define TIMEOUT 10 /*when we are registering, the FA waits TIMEOUT seconds
                    b4 deletes the entry in fnodes. It can be changed*/
#define LENGTHMD5 16
#define MULTICAST "224.0.0.1"   /*Predefined multicast adddress*/
#define DATAFILE "configuration"/*configuration file*/

#define RFLAG 0x8000   /*FA Registration bit*/
#define BFLAG 0x4000   /*Busy bit*/
#define	HFLAG 0x2000   /*Agent offers service as home agent*/
#define FFLAG 0x1000   /*Agent offers service as foreign agent*/
#define MFLAG 0x0800   /*Minimal encapsulation*/
#define GFLAG 0x0400   /*GRE encapsulation*/
#define VFLAG 0x0200   /*Agent supports Van Jacobson header compression*/

/*#define ANOTHER_MOBILE_NODE
This option, if activated, is used to create a permanent entry in the
ARP cache. If the implementation for the MN is the same as the implementation
of this FA, this option could be desactivated*/


struct FAvalues ForeignA;
struct HAvalues HomeA;
struct MNode mynodes[10];
struct MNode fnodes[10];
char packet[MAX_PACKET];
int SockAgentInOut;
unsigned char errorvalue;
struct authentication *authen;
char interface[5];

/********************************************************************/
/********************************************************************/
/********************************************************************/

#include "parameters.h"
#include "arpissues.h"
#include "route-arp-ag.h"

void initsockets()
{

  /*This procedure is in charge of creating an UDP socket*/

   struct sockaddr_in sa;
   int enable = 1, value;
   struct protoent *proto;
   struct ip_mreq multicast;
   
   SockAgentInOut = socket(AF_INET, SOCK_DGRAM, 0);

   if (SockAgentInOut < 0)  
   	{
	  fprintf(stderr, "initsockets(): could not create sockets.\n");
	  exit(-1);
	}
              
   sa.sin_family = AF_INET; 
   sa.sin_addr.s_addr = htonl(INADDR_ANY); 
   sa.sin_port = htons(AGENT_LISTEN_PORT); 
   
   if (bind(SockAgentInOut, (struct sockaddr *) &sa, sizeof(sa)) < 0)  {
      perror("initsockets(): Bind SockAdvertInput failed.\n");
      exit(-1);
   }

DEBUG (4) 
{
  printf ("Sockets initializated\n");
  printf ("RAW socket for adverts and normal sockets to registrarion rep and req\n");
}   
DEBUG (0) printf ("HA/FA capabilities are working\n");
}

/********************************************************************/

void initall()
{

  /*This procedure is in charge of setting up all the values*/

unsigned char temp[20];
int i;

loadparameters(DATAFILE);
initsockets();
ForeignA.seq_num=0;
ForeignA.Lifetime=LIFETIME; 
ForeignA.TimeBetweenAdverts = 2; /*Time between two adverts. It can be changed*/
ForeignA.flags = HFLAG | FFLAG | RFLAG;
HomeA.Lifetime = LIFETIME;
HomeA.NetMask = 0xffff0000;

DEBUG(4)
printf ("HA's MAC address: %x:%x:%x:%x:%x:%x\n", HomeA.hwAddress[0], HomeA.hwAddress[1], 
			       HomeA.hwAddress[2], HomeA.hwAddress[3],
                               HomeA.hwAddress[4], HomeA.hwAddress[5]);

for (i=0; i<ForeignA.supported; i++)
  {
     fnodes[i].status = 0;
     fnodes[i].index = 0;
  }

for (i=0; i<HomeA.supported; i++)
  {
     fnodes[i].status = 0;
     fnodes[i].index = i;
  }

startTimers();
/*sAgentAdvert();
*/
}
/********************************************************************/

int checkChecksumICMP (struct icmphdr *icmp)
{
unsigned short swapvar;

swapvar=icmp->checksum;
icmp->checksum = 0;
icmp->checksum = in_cksum (icmp, LENGTHAGENTADVERT);
if (swapvar != icmp->checksum) return 0;
return 1;

}
/********************************************************************/

void sendRegResponse (unsigned long Address, int position, int select_time)
{

  /*This function is used by the home agent. It sends a registration response 
    to the mobile node*/

int value, sent;
struct MIPRegRep *Response;
struct sockaddr whereto;
struct sockaddr_in *to;
unsigned long temp;


bzero ((char *) &whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;
to->sin_addr.s_addr =  Address;
to->sin_port = mynodes[position].port; /*htons(434);*/

Response = (struct MIPRegRep *)packet;
Response->type = 3;
Response->code =  errorvalue;
/*Change made for Page 52 for deregisterng with a previous FA*/
if (select_time)
    Response->lifetime = htons(mynodes[position].lifetime); 
else Response->lifetime = 0;
/*End of change*/
Response->Address = mynodes[position].Address;
Response->homeAgent = mynodes[position].homeAgent;
Response->identification[0] = htonl (mynodes[position].identification[0]);
Response->identification[1] = htonl (mynodes[position].identification[1]);

/*HA-MN authentication */

value = sizeof (struct MIPRegRep);

*(packet +value++) = 32;
*(packet + value++) = 4 + LENGTHMD5;
temp = mynodes[position].SPI;
bcopy (&temp, packet+value, 4);
appendauth (packet, value, mynodes[position].MD5key, mynodes[position].MD5keylength);
value += 4 + LENGTHMD5;


if (authenAssociation (Address, authen))
  {
     struct authentication *position;
     unsigned long temp;

     position = authenAssociation (Address, authen);
     *(packet + value++) = 34;
     *(packet + value++) = 4 + LENGTHMD5;
     temp = position->SPI;
     bcopy (&temp, packet+ value, 4);
     appendauth (packet, value, position->MD5key, position->MD5keylength);
     value +=  4 + LENGTHMD5; 
  }

sent = sendto (SockAgentInOut, (char *) packet, value, 0, &whereto, sizeof (struct sockaddr)); 
if (sent != value) perror ("sendRegResponse() sendto error");
DEBUG(3) printf ("I've sent a packet in response to a registration packet (to the MN)\n");
}

/*******************************************************************/

void nodeIsOff (void *position)
{

/*Used by Home Agent. This function is called when the home agent does not
receive a new registration request, therefore the home agent supposes that
the mobile node is off and it removes information about that mobile node*/

int *index;

DEBUG (4) printf ("Node is off ...");
index = (int *) position;
mynodes[*index].status = 0; /*MN has not done a registration request*/
mynodes[*index].careOfAddress = (unsigned long) 0;
DEBUG (5) printf ("Entry of the removed node: %d\n", *index);
delARPentry (mynodes[*index].Address, 1);
routeEntry (2, mynodes[*index].Address, 0, inet_addr("255.255.255.255"));
delIFtunl (*index);
}

/*******************************************************************/

void saveNodeData (char *packet, int length, int position, struct sockaddr_in from, int remote)
{

/*Used by the home agent. This function is used to store information from
a registration request. It checks that the registration request is valid*/

/*remote: This value is set to 1 if the MN is away from home but is registering with
the home agent. If the MN is registering with the home agent but it's on its home
network, remote is equal to 0*/

struct MIPRegReq *Req;
int extensions;
unsigned long previousAddress;

/*It must be checked this is a UDP packet (valid, I mean!)*/

Req = (struct MIPRegReq *)packet;
if (Req->type != 1) return;

/*This is not a registration packet. Something must be changed, because when
this procedure fishishes a packet is sent to the MN. The packet received
must be ignored, therefore no packet must be sent*/

errorvalue = 1;

if (ntohs(Req->flags))
  {
     /*At the moment, flags must be 0*/
     errorvalue = 130;
  }
mynodes[position].port = /*htons(AGENT_LISTEN_PORT); */from.sin_port;
previousAddress = mynodes[position].careOfAddress;
mynodes[position].careOfAddress = Req->careOfAddress;
mynodes[position].identification[0] = ntohl(Req->identification[0]);
mynodes[position].identification[1] = ntohl(Req->identification[1]);
DEBUG (3)
  printf ("Identification %u\n",ntohl(Req->identification[1]));
mynodes[position].homeAgent = Req->homeAgent;
mynodes[position].Address = Req->Address;
mynodes[position].lifetime = ntohs(Req->lifetime) > HomeA.Lifetime ? HomeA.Lifetime : ntohs(Req->lifetime);
mynodes[position].index = position;

/*length will be used to know where HA-FA extensions start and it will be saved in mynodes*/

if (length > sizeof (struct MIPRegReq))
  {
    extensions = sizeof (struct MIPRegReq);
    while ((*(packet +extensions) != 32) && (extensions < length))
      extensions += 2 + *(packet + extensions +1);
    if (extensions >= length) errorvalue = 134; /*Error: poorly formed request*/
    if (*(packet + extensions) == 32)
     {
        /*MN-HA authentication extension*/
	unsigned char MD5keyfound[KEYLENGTH];
	unsigned long *SPI;
	bcopy ((packet + extensions + 6), MD5keyfound, LENGTHMD5);
        appendauth(packet,extensions +2 ,mynodes[position].MD5key,mynodes[position].MD5keylength );
        
	SPI = (unsigned long *)(packet + extensions + 2);
        if ((*SPI >= 0) && (*SPI <= 255))
          {
		DEBUG (1) printf ("SPI error\n");
		errorvalue = 128;
	  }
        if (bcmp(packet +extensions + 6, MD5keyfound, LENGTHMD5 ))
           {
             DEBUG (0) printf ("Error: Authentication between MN-HA not valid\n"); 
             errorvalue = 131; /*MN failed authentication*/
           }
     }

    if ((remote) && (errorvalue == 1) && (authenAssociation(mynodes[position].careOfAddress,authen)))
     {
       /*Authentication between FA & HA*/

       while ((*(packet +extensions) != 34) && (extensions < length))
        extensions += 2 + *(packet +extensions +1);
       if (extensions >= length) errorvalue = 132; /*FA-HA authenticator MUST be present*/
       if (*(packet +extensions) == 34)
        {
          /*MN-HA authentication extension*/
          unsigned char MD5keyfound[KEYLENGTH];
          struct authentication *foreignKey;
	  unsigned long *SPI;

          foreignKey = authenAssociation (mynodes[position].careOfAddress,authen);
	  
          SPI = (unsigned long *)(packet + extensions +2);
	  if ((*SPI >= 0) && (*SPI<= 255))
             {
		DEBUG (1) printf ("SPI error\n");
		errorvalue = 128;
	     }
          bcopy ((packet + extensions +6),MD5keyfound, LENGTHMD5);
          appendauth (packet, extensions + 2, foreignKey->MD5key, foreignKey->MD5keylength);
          if (bcmp(packet + extensions + 6, MD5keyfound, LENGTHMD5))
           {
             DEBUG (0) printf ("Error: Authenticatioin between FA-HA not valid\n");
             errorvalue = 132; /*FA failed authentication*/
           }
        } 

     }

  }
else errorvalue = 134; /*There is no MN-HA authenticator available*/

/*Now It creates a timeout for those nodes which are away from home.
If a node is at home and it's registering now, this timeout is not activated*/

if (((errorvalue == 1 ) || (!errorvalue)) && (mynodes[position].status))
    if (mynodes[position].status) cancelTimer (mynodes[position].timeout1); 

if (((errorvalue == 1) ||  (errorvalue == 0)) && (remote) && (mynodes[position].lifetime))
	mynodes[position].timeout1 = 
        addTimer((mynodes[position].lifetime+mynodes[position].lifetime/8), nodeIsOff,(void *)&mynodes[position].index, NULL );   
if (errorvalue == 1) DEBUG (3) printf ("%s packet accepted\n",mynodes[position].status ? "Registration" : "Deregistration");

if (((!remote) || ((remote) && (!mynodes[position].lifetime))) && (errorvalue == 1))
  {
    if ((mynodes[position].status) && (previousAddress))
     {
       /*Mn was away and now it is at home*/
       delARPentry(mynodes[position].Address, 1); /*It deletes proxy arp*/
       delIFtunl(position); /*Delete tunnel interface*/
       DEBUG(2) printf ("If an error comes just after this comment, ignore it!\n");
       routeEntry (2, mynodes[position].Address, 0,(unsigned long)inet_addr("255.255.255.255")); /*Delete route to MN*/
       /*Probably, HA must have an entry as follows: MN -> dev0.
         This is necessary because when the MN is at home, and HA was
         acting as a proxy arp, some arp caches are updated with HA's MAC
         address. So, the correspondents hosts will send packets to the
	 HA*/
       if ((authenAssociation(previousAddress, authen)) && (remote) && (!mynodes[position].lifetime)) /*Node coming back home
	   and there is a binding assotiation between HA and previous FA*/
	  /*Then it sends a deregistration request to the previous FA (see pg 53 RC2002)*/
	  printf ("It got this way\n");
          sendRegResponse (previousAddress, position, 0);
     }
    mynodes[position].status = 0;
  }
else if ((remote) && (errorvalue == 1) && (previousAddress != HomeA.Addr)
          && (previousAddress != mynodes[position].careOfAddress) 
          && (previousAddress)) 
       {
         changeTunl(position, mynodes[position].careOfAddress);
         mynodes[position].status = 1;
       }
     else if ((errorvalue == 1) && (previousAddress != mynodes[position].careOfAddress))
            {
		/*MN is on a Forign Network for the first time!*/
                char device[20];
                device[sprintf(device, "tunl%c", position+'0')] = '\0';
                DEBUG (2) printf ("Creating tunnel %s\n", device);
                addIFtunl (position, mynodes[position].careOfAddress); 
		DEBUG (2) printf ("Creating a proxy ARP entry\n");

		/*Next instrucction MUST be executed. Otherwise, if our agent was down and
		we run it again without clear routing table, it will not work. See kernel*/

                routeEntry (2, mynodes[position].Address, 0,(unsigned long)0);
		delARPentry (mynodes[position].Address, 0);
		proxyARP (mynodes[position].Address, HomeA.hwAddress);
                DEBUG (2) printf ("Creating an entry on the routing table\n");
		routeEntryDev(mynodes[position].Address, device, inet_addr ("255.255.255.255")); 
		GratuitousARP (mynodes[position].Address, HomeA.hwAddress);
		addTimer (2, gratARPbackground, (void *)&mynodes[position].Address, (void *)HomeA.hwAddress);
		addTimer (5, gratARPbackground, (void *)&mynodes[position].Address, (void *)HomeA.hwAddress);
		/*gratuitousARP (mynodes[position].Address, HomeA.hwAddress);*/ 
                mynodes[position].status = 1;
            }


/*If it is a re-registration request, it does not change routing tables, arp entries, ... */

if ((errorvalue != 1) && (errorvalue != 0) && (mynodes[position].status))
  {
     delARPentry (mynodes[position].Address, 1);
     delIFtunl (position);
     routeEntry (2, mynodes[position].Address, 0, inet_addr ("255.255.255.255"));
     mynodes[position].status = 0;
  }
}

/*******************************************************************/

void noRegist (void *arg )
{

  /*This procedure is used by the foreign agent. When this function is called
    it is because the foreign agent does not have obtained another registration
    request. It assumes that the mobile node is not on its subnet so, it
    cancels foreign agent's capabilities to that mobile node*/

  int *index;
  index = (int *)arg;

  routeEntry (2, fnodes[*index].Address, 0, inet_addr ("255.255.255.255"));
  delARPentry (fnodes[*index].Address, 0);
  fnodes[*index].status = 0;
}

/*******************************************************************/

int wasWaiting (unsigned long Address)

{

  /*This function returns != -1 if the mobile node was waiting for a registra
    tion reply and -1 otherwise. Used by the foreign agent*/

 int i;
 for (i=0;i<ForeignA.supported;i++)
   if ((fnodes[i].Address == Address) && (fnodes[i].status == 2)) 
          return i;
 return -1;
}

/*******************************************************************/

void sendingerrors (char *packet, int length, struct sockaddr *to)
{

  /*This function is used to send a reply to the mobile node. Usually, when
    an error has been found.*/

int bytes;

bytes = sendto (SockAgentInOut, (char *)packet, length, 0, to, sizeof (struct sockaddr));
if (bytes != length) DEBUG (0) printf ("sendingerrors() sendto error"); 
}

/*******************************************************************/

void passRepOn (char *packet, int length, unsigned long Address)
{

/*Used by the foreing agent. It passes the registration reply on to the mobile 
  node.*/

int place, bytes, extensions, tempsock;
struct sockaddr whereto;
struct sockaddr_in *to;
struct MIPRegRep *Response;

Response = (struct MIPRegRep *)packet;

bzero((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;
to->sin_addr.s_addr = Address;

if ((Response->lifetime == 0) && (wasConnected(Response->Address != -1)))  
{
extensions = sizeof (struct MIPRegRep);

if (authenAssociation(Response->homeAgent, authen))
 {
   while ((*(packet +extensions) != 34) && (extensions < length))
      extensions += 2 + *(packet + extensions + 1);
 
   if (extensions >= length)
     {
       Response->code = 132;
       sendingerrors (packet, length, &whereto);
       return;
     }
   if (*(packet + extensions) == 34)
    {
      unsigned char MD5keyfound[KEYLENGTH];
      struct authentication *index;
      unsigned long *SPI;

      index = authenAssociation (Response->homeAgent, authen);
      SPI = (unsigned long *)(packet + extensions +2);
      if ((*SPI >= 0) && (*SPI <=255))
       {
         DEBUG (1) printf ("SPI error \n");
         Response->code = 64; /*SPI are differents!*/
         sendingerrors (packet, length, &whereto);
         return;
       }

      bcopy((packet + extensions + 6), MD5keyfound, LENGTHMD5); 
 
      appendauth (packet, extensions + 2, index->MD5key, index->MD5keylength);
      if (bcmp(packet + extensions + 6, MD5keyfound, LENGTHMD5))
        {
          DEBUG (0) printf ("Authentication between HA-FA not valid\n");
          Response->code = 131; /*errorvalue*/
          sendingerrors (packet, length, &whereto);
          return;
        }
     }
 }
DEBUG (2) printf ("Deregistration packet valid from HA.\n");
place = wasConnected(Response->Address);
cancelTimer (fnodes[place].timeout1);
noRegist ((void *)&place);
return;
}

if (place = wasWaiting(Response->Address) == -1)
  {
     DEBUG (1) 
       {
        printf ("FA's got an registration response from HA\n");
        printf ("when a registration request has not been issued\n");
       }
     return;
  }
else DEBUG (1) printf ("FA was waiting for this reply. Thanks!\n");

to->sin_port = fnodes[place].port;
cancelTimer (fnodes[place].timeout1);

if ((Response->code != 0) && (Response->code != 1))
  {
     DEBUG (1) printf ("HA has denied the registration request!. Response code %d\n",Response->code);
     sendingerrors (packet, length, &whereto);
     return;
  }
else DEBUG (1) printf ("The HA has accepted the registration request ... good!\n");

extensions = sizeof (struct MIPRegRep);

if (authenAssociation(Response->homeAgent, authen))
 {
   while ((*(packet +extensions) != 34) && (extensions < length))
      extensions += 2 + *(packet + extensions + 1);
  
   if (extensions >= length) 
     {
       Response->code = 132;
       sendingerrors (packet, length, &whereto);
       return;
     }
   if (*(packet + extensions) == 34)
    {
      unsigned char MD5keyfound[KEYLENGTH];
      struct authentication *index;
      unsigned long *SPI;

      index = authenAssociation (Response->homeAgent, authen);
      SPI = (unsigned long *)(packet + extensions +2);
      if ((*SPI >= 0) && (*SPI <=255))
       {
	 DEBUG (1) printf ("SPI error \n");
	 Response->code = 64; /*SPI are differents!*/
         sendingerrors (packet, length, &whereto);
	 return;
       }

      bcopy((packet + extensions + 6), MD5keyfound, LENGTHMD5); 
 
      appendauth (packet, extensions + 2, index->MD5key, index->MD5keylength);
      if (bcmp(packet + extensions + 6, MD5keyfound, LENGTHMD5))
        {
          DEBUG (0) printf ("Authentication between HA-FA not valid\n");
          Response->code = 131; /*errorvalue*/
	  sendingerrors (packet, length, &whereto);
	  return;
        }
     }
 }

DEBUG (1) printf ("ok! Authentication test ... checked!\n");

extensions = sizeof (struct MIPRegRep);
if (authenAssociation (Response->Address, authen))
 {
    struct authentication *index;
    unsigned long temp;


    while ((*(packet + extensions) != 32) && (extensions < length))
      extensions += 2 + *(packet + extensions + 1);
    if (*(packet + extensions) == 32) extensions += 2 + *(packet + extensions +1);

    /*Now, authentication between MN and FA must be inserted*/

    index = authenAssociation (Response->Address, authen);
    *(packet +extensions++) = 33; /*MN-FA authentication*/
    *(packet + extensions++) = 4 + LENGTHMD5;
    temp = index->SPI;
    bcopy (&temp, packet + extensions, 4);
    appendauth (packet, extensions, index->MD5key, index->MD5keylength);
    length = extensions +4 + LENGTHMD5;
 }

routeEntryDev (Response->Address,interface,inet_addr ("255.255.255.255") ); 
fnodes[place].status = 1; /*UP*/
fnodes[place].timeout1 = addTimer (ntohs(Response->lifetime), noRegist, (void *)&fnodes[place].index, NULL);
#ifdef ANOTHER_MOBILE_NODE
	createARPPerm (Response->Address);
#endif
tempsock = socket (AF_INET, SOCK_DGRAM,0);
bytes = sendto(tempsock/*SockAgentInOut*/, (char *)packet, length, 0, &whereto, sizeof(whereto));
close (tempsock);
if (bytes < 0 || bytes != length) perror ("passRepOn() sendto error");
}

/*******************************************************************/

int lookingplace (unsigned long Address)
{
  /* this function returns != -1 if it can support a mobile node. Otherwise, 
     it returns -1. It is used by the foreign agent.*/
  int i;

  for (i=0; i< ForeignA.supported; i++)
    if ((fnodes[i].Address == Address) && (fnodes[i].status))
         return i;

  for (i= 0; i<ForeignA.supported;i++)
   if (!fnodes[i].status) return i;

  errorvalue = 66; /*Insufficient resources*/
  return -1;
}

/*******************************************************************/

int wasConnected (unsigned long Address)
{
  /*This function returns -1 if a mobile node was not connected to this agent.
    Otherwise it returns the possition of the mobile node in the array of 
    foreign mobile nodes. It is used by the foreign agent*/

int i;

for (i=0;i<ForeignA.supported;i++)
  if ((fnodes[i].Address == Address) && (fnodes[i].status == 1))
      return i;
return -1;
}

/*******************************************************************/

void sendPacketToMNFromFA (char *packet, int length, struct sockaddr_in from)
/*It's used just when an error has been detected (by the FA) and it sends the
error to the mobile node.*/

{

struct MIPRegReq *Request;
struct MIPRegRep *Res;
struct sockaddr whereto;
struct sockaddr_in *to;

char response[1024];
int ind, sent, value;

Request = (struct MIPRegReq *) packet;
bzero ((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET ;
to->sin_addr.s_addr = Request->Address; 
to->sin_port = /*htons (434);*/ from.sin_port;   

ind = sizeof (struct MIPRegRep);
Res = (struct MIPRegRep *)response;

Res->type = 3;
Res->code = errorvalue;
Res->lifetime = htons(LIFETIME);
Res->Address = Request->Address;
Res->homeAgent = Request->homeAgent;
Res->identification[1] = ntohl(Request->identification[1]);
Res->identification[0] = ntohl(Request->identification[0]);

if (authenAssociation(Res->Address,authen))
    /*Put authentication between MN and FA*/
   {
      unsigned long temp;
      struct authentication *place;

      place = authenAssociation (Res->Address, authen);
      *(response + ind++) = 33; /*MN-FA authen*/
      *(response + ind++) = 4 + LENGTHMD5;
      temp = place->SPI;
      bcopy (&temp, response + ind , 4);
      appendauth ( response, ind , place->MD5key, place->MD5keylength);  
      ind += 4 + LENGTHMD5;
   }

if (value = wasConnected(Request->Address) != -1)
   {
     routeEntry (2, Request->Address, 0, inet_addr ("255.255.255.255"));
     fnodes[value].status = 0;
   }

DEBUG(1)
 {
    printf ("Error: A packet has been sent to the MN from the FA. ");
    printf ("Reason %d\n",errorvalue);
 }

routeEntryDev (Request->Address, interface,inet_addr ("255.255.255.255")); 

sent = sendto (SockAgentInOut, response, ind, 0, &whereto, sizeof (struct sockaddr));
if (sent != ind) perror ("sendPacketFromMNFromFA() sendto error");

routeEntry (2, Request->Address, 0, inet_addr("255.255.255.255"));
}

/*******************************************************************/

void didntArrive (void *index)
{

  /*This procedure is called when a registration reply has not been received
    by the foreign agent.*/

int *pointer;

pointer = (int *)index;
fnodes[*pointer].status = 0;

/*It has nothing to do because this function is called when a Registration reply
is not sent from the HA. So, It removes its entry in fnodes array, setting 
status equal to 0*/
}

/*******************************************************************/


void passRegOn (char *packet, int length, unsigned long Address, struct sockaddr_in from)
{

  /*This procedure is used by the foreign agent. It passes a registration reply
    on to the home agent.*/

struct MIPRegReq *Request;
struct sockaddr whereto;
struct sockaddr_in *to;
int index, ind, sent ;

Request = (struct MIPRegReq *)packet;

/*Looking for place in fnodes array*/

index = lookingplace (Request->Address);
if (index == -1)
  {
     DEBUG(1) printf ("FA cannot admit more MN\n");
     sendPacketToMNFromFA(packet, length, from);
     return;
  }

if (ntohs(Request->lifetime) > ForeignA.Lifetime) 
    {
      DEBUG(1) printf ("Error: Lifetime denied by the FA. Lifetime too long\n");
      errorvalue = 69; 
      sendPacketToMNFromFA (packet, length,from);
      return;
    }

fnodes[index].Address = Request->Address;
fnodes[index].homeAgent = Request->homeAgent;
fnodes[index].lifetime = ntohs (Request->lifetime);
fnodes[index].identification[0] = ntohl(Request->identification[0]);
fnodes[index].identification[1] = ntohl(Request->identification[1]);
fnodes[index].index = index;

fnodes[index].port = from.sin_port;

if (fnodes[index].status)
    cancelTimer (fnodes[index].timeout1);

if ((ind = sizeof (struct MIPRegReq)) < length)
 {
   while ((*(packet + ind) != 32) && (ind < length))
       ind += 2 + *(packet +ind +1);
   if (ind >= length)
     {
       DEBUG (1) printf ("Error: MN-HA authentication is not available\n");
       errorvalue = 70;       /*poorly formed request*/
       printf ("length %d\n",length);
       sendPacketToMNFromFA (packet, length, from);
       return;
     } 
   if (authenAssociation(Request->Address, authen))
    {
      while ((*(packet +ind) != 33) && (ind < length))
          ind += 2 + *(packet + ind + 1);
      if (ind >= length)
         {
             DEBUG (1) printf ("Error: MN-FA authentication should be here\n");
             errorvalue = 67; /*MN failed authentication*/
	     sendPacketToMNFromFA (packet, length, from);
             return;
          }
      else
         {
           unsigned char MD5keyfound[KEYLENGTH];
           unsigned long *SPI;
           struct authentication *temp;

	   temp = authenAssociation (Request->Address, authen); 
           SPI = (unsigned long *)(packet+ind+2);
	   if ((*SPI >= 0) && (*SPI <=255))
             {
                DEBUG (1) printf ("SPI error\n");
		errorvalue = 64;
                sendPacketToMNFromFA (packet, length, from);
                return;
             }
           bcopy (packet + ind + 6, MD5keyfound, LENGTHMD5);
           appendauth(packet, ind + 2, temp->MD5key, temp->MD5keylength);
           if (bcmp(MD5keyfound, packet +ind + 6, LENGTHMD5))
            {
	      DEBUG (1) printf ("MN-FA authentication not valid\n");
              errorvalue = 67;
              sendPacketToMNFromFA (packet, length, from);
              return;
            }       
         }
    }
 }
else
 {
   DEBUG (1) printf ("Error: MN-HA authentication is not available\n");
   errorvalue = 70;    /*poorly formed request */
   printf ("Size of struct %d, length %d\n", sizeof (struct MIPRegReq), length);
   sendPacketToMNFromFA (packet, length, from);
   return;
 }

/*Packet checked out. No problems detected so it must be sent to correspondent HA.
Creating packet*/

DEBUG(1) printf ("Packet accepted\nCreating a request to HA\n");

fnodes[index].status = 2; /*WAITING*/

ind = sizeof (struct MIPRegReq);
if (authenAssociation(Request->homeAgent, authen))
  {
    unsigned long temp;
    struct authentication *intermediate;

    ind = sizeof (struct  MIPRegReq);

    while ((*(packet +ind) != 32) && (ind < length))
      ind += 2 + *(packet + ind + 1);
    
    intermediate = authenAssociation(Request->homeAgent, authen);

    ind += 2 + *(packet + ind + 1);
    *(packet + ind++) = 34;
    *(packet + ind++) = 4 + LENGTHMD5;
    temp = intermediate->SPI;
    
    bcopy (&temp, packet + ind, 4);
    appendauth (packet, ind, intermediate->MD5key, intermediate->MD5keylength);
    ind += 4 + LENGTHMD5;
  }
else ind = length;

bzero ((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;
to->sin_addr.s_addr = Request->homeAgent;
to->sin_port = htons (AGENT_LISTEN_PORT);

sent = sendto (SockAgentInOut, (char *)packet, ind, 0, &whereto, sizeof (struct sockaddr));
if (sent != ind) perror ("passRegOn() sendto error"); 
else 
  {
     DEBUG (1) printf ("Packet sent to HA successfully\n");
     fnodes[index].timeout1 = addTimer (TIMEOUT, didntArrive, (void *)&fnodes[index].index, NULL);
  }
}

/*******************************************************************/

void checkRegPacket (char *packet, int length, struct sockaddr_in from)
{

  /*This function is used by both the home agent and foreign agent. It checks
    that a registration packet (that has been received) is correct and it 
    identifies where it comes from. Depending on that, it is acting as foreign
    agent or home agent.*/

   struct MIPRegReq *Request;
   struct MIPRegRep *Reply;

   Request = (struct MIPRegReq *)packet;
   Reply = (struct MIPRegRep *)packet;
   DEBUG(4)
    {
     printf ("I'VE GOT A PACKET FROM THE MN IN ORDER TO REGISTER IT\n");
     printf ("Type: %d flags %d Lifetime %d Home Address %s\n",Request->type, ntohs(Request->flags),
              ntohs(Request->lifetime), inet_ntoa(Request->Address)); 
     printf ("Home Agent %s \nIdentification0 %d Identification1 %d\n",
              inet_ntoa(Request->homeAgent), ntohl(Request->identification[0]), 
              ntohl(Request->identification[1]));
     printf ("Care of Address %s\n",inet_ntoa(Request->careOfAddress));
    }
   DEBUG (3)
     printf ("This packet comes from %s\n",inet_ntoa(from.sin_addr.s_addr));
   if ((belongstome(Request->Address)) &&(Request->Address == from.sin_addr.s_addr))/*Acting as HA*/
     {
        /*The node belongs to this network and it's at home, so lifetime must be 0*/
        if (ntohs(Request->lifetime)) 
         {
           DEBUG(0) printf ("Error: Lifetime MUST be 0\n");
           errorvalue = 69; 
           sendRegResponse (Request->Address,whichmynode(Request->Address), 1); 
           return;
         }
       saveNodeData (packet, length, whichmynode(Request->Address),from,0); 
              /*'whichmynode' says the node position in the MNode array*/
       sendRegResponse (Request->Address, whichmynode(Request->Address), 1); 
                 /*if errorvalue = 0 or 1, there is not error*/
     } 
   else if (belongstome(Request->Address))
    {
      /*It's acting as HA but MN is away from home*/
 
      saveNodeData (packet, length, whichmynode(Request->Address),from, 1);
      sendRegResponse (Request->careOfAddress, whichmynode (Request->Address), 1);
    }
        else if (Request->Address == from.sin_addr.s_addr) /*It's acting as FA.MN sends to him a reg. req*/
	       {
                 passRegOn (packet, length, Request->homeAgent,from);
	       }
             else  /*HA has sent a registraton reply to the MN*/
                    {
			  DEBUG (3) printf ("I've got a packet from the HA\n");
			  passRepOn (packet, length, Request->Address);
                    }
}


/********************************************************************/
/********************************************************************/
/********************************************************************/
/********************************************************************/


main(int argc, char *argv[])
{
      fd_set fdvec, fdvec2;
      struct timeval tv;
      int selcode;

      int packlen=MAX_PACKET;
      struct sockaddr_in from;
      int length, fromlen;
      
      fromlen = sizeof (from);
      if (argc ==2)
	   debugLevel = atoi(argv[1]);
      initall();           /*setting up values */
  for(;;)
  {
        if ((length = recvfrom (SockAgentInOut, (char *)packet, packlen, 0,
	    (struct sockaddr *)&from, &fromlen)) < 0)
         {
           if (errno == EINTR) continue;
           DEBUG (4) 
	   { 
		perror ("main() recvfrom error");
           }
	  continue;
         }
        DEBUG (4) printf ("I've got an UDP packet\n");
        checkRegPacket (packet, length, from);
      }
} 

/********************************************************************/

int whichmynode (unsigned long Address)
{
int i;

for (i=0;i<HomeA.supported;i++)
  if (mynodes[i].Address == Address) return i;

return -1;

}

/********************************************************************/

int belongstome (unsigned long Address)
{
  if (whichmynode(Address) >= 0) return 1;
  return 0;
}

 
 
