/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/

/*This module is the deamon which once is compiled is running on the mobile
host*/

#define ARP_REPLIES

/*If this define is activated, it means that the MN sends ARP replies to the
FA in order to mantain updated its ARP cache. If it's not active, it implies
that the FA is able to obtain MN's MAC address by other means i.e. (from the registra
tion request and making it permanent)*/


#include "includes_mn.h"

#define AGENT_LISTEN_PORT 434            /*Predefined port*/
#define MAX_PACKET 5000
#define LENGTHAGENTADVERT 8
#define AGENTSOLEVERY 5                  /*time between agent solicitaions*/
#define WAITINGANSWER  5                 /*It waits for a registration reply*/
#define INTERFACE "eth0"                 /*The "mobile" interface*/
#define MULTICAST "224.0.0.1"            /*Predefined multicast address*/
#define DATAFILE "configuration"         /*Configuration file*/
#define DATAFILE2 "routeentry.conf"      /*default routes*/
#define LENGTHMD5 16
#define MAX_ADVERT_TIME 60		/*Max time between two Agent solicitation requests*/
/*#define SEVERAL_AGENTS_OPTION
 */

/*If this define is activated, the mobile nodes supposes that more than one
agent is working on the same network. If it is not activated, it means there
is just one agent (foreign agent) per network. It implies that if just agent is
active, as soon as It sees a new agent advertisement different to the previous
one, it will attempt to register.*/

#define DOWN 0
#define WAITING 1
#define UP 2

#define RFLAG 0x8000   /*FA Registration bit*/
#define BFLAG 0x4000   /*Busy bit*/
#define	HFLAG 0x2000   /*Agent offers service as home agent*/
#define FFLAG 0x1000   /*Agent offers service as foreign agent*/
#define MFLAG 0x0800   /*Minimal encapsulation*/
#define GFLAG 0x0400   /*GRE encapsulation*/
#define VFLAG 0x0200   /*Agent supports Van Jacobson header compression*/

void changeState();
struct MIPAssocState *stateChain;
char packet[MAX_PACKET];
int idSol; /*Identification of the Solicitations in timer.c*/
int seed; /*Used to help backoff algorithm for agent solicitations*/
struct mobileNode mynode;
int SockAgentInOut, SockAdverts; 
unsigned long tempCOA;
char interface[5];
#include "route-arp-mn.h"
#include "arpissues.h"

/********************************************************************/
/********************************************************************/
/********************************************************************/

void initsockets()
{
  /*It creates two sockets:
    1) to send and receive agent solicitations/advertisements
    2) to send UDP registration requests and to receive UDP registrations
       replies.
       */

   struct sockaddr_in sa;
   int enable = 1;
   struct protoent *proto;
   struct ip_mreq mreq;

   if ((proto = getprotobyname("icmp")) == NULL) 
   {
                fprintf(stderr, "unknown protocol icmp\n");
                exit(1);
   }

   SockAgentInOut = socket(AF_INET, SOCK_DGRAM, 0);
   SockAdverts = socket(AF_INET, SOCK_RAW, proto->p_proto);
  
   if (SockAgentInOut < 0 || SockAdverts  < 0 )  
   	{
	  fprintf(stderr, "initsockets(): could not create sockets.\n");
	  exit(-1);
	}
              
   sa.sin_family = AF_INET; 
   sa.sin_addr.s_addr = htonl(INADDR_ANY); 
   sa.sin_port = htons(AGENT_LISTEN_PORT); 
   
   if (bind(SockAgentInOut, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) < 0)  
   {
      perror("initsockets(): Bind SockAdvertInput failed.\n");
      exit(-1);
   }

   if (setsockopt(SockAdverts, SOL_SOCKET, SO_BROADCAST, 
		  (char *)&enable, sizeof(int)) < 0)  
       {
	  perror("initsockets(): setsockopt SO_BROADCAST failed");
	  exit(-1);
       }

   /*MULTICASTING*/
   bzero ((char *)&mreq, sizeof (struct ip_mreq));
   mreq.imr_multiaddr.s_addr = inet_addr (MULTICAST);
   mreq.imr_interface.s_addr = mynode.Address;

   if (setsockopt(SockAdverts, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(struct ip_mreq))<0 )
    {
      DEBUG (0) printf ("Error: it cannot join multicast group!\n");
      exit (-1);
    }

DEBUG (4) 
{
  printf ("Sockets initializated\n");
  printf ("RAW socket for adverts and normal socket to registrarion rep and req\n");
}   
}

/********************************************************************/

void sendSol()
{

  /*This function is called when the mobile node has to send agent solicitati
    ons */

char buffer[1024];
struct icmphdr  *icp;
int i, length; 
struct sockaddr whereto;
struct sockaddr_in *to;

bzero ((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;
to->sin_addr.s_addr = htonl (INADDR_BROADCAST); /*inet_addr (MULTICAST);*/


bzero (buffer,1024);
icp = (struct icmphdr *)buffer;
icp->type = 10;

length = 8; /*type(1 byte)+ code(1 byte)+ Checksum(2 bytes)+ reserved(4 bytes)*/

/* compute ICMP checksum here */

icp->checksum = in_cksum((u_short *)buffer, length);

i = sendto(SockAdverts, (char *)buffer, length, 0, &whereto, sizeof (struct sockaddr));
if (i < 0 || i != length) perror("sendSol () sendto error");
else
DEBUG(1) 
  printf ("Agent Solicitation packet sent\n");
}

/********************************************************************/

int calculateTime ()
{
static int time = 1;

switch (seed)
{
	case 0:
		time = 1;
		seed++;
		break;
	default:
		if (seed > 2) time+= time;
		else seed++;
		if (time > MAX_ADVERT_TIME)
			time = MAX_ADVERT_TIME;
		DEBUG (2) printf ("Time of next advert %d. Max time %d\n",
				time, MAX_ADVERT_TIME);
                
		break;
}

return time;
}

/********************************************************************/

void sAgentSol ()
{


sendSol();
idSol = addTimer (calculateTime(), sAgentSol, NULL,NULL);
}

/********************************************************************/

void initall()
{

  /*It initialites all the data in the mobile node*/

readValuesFromFile(DATAFILE); /*it reads the configuration file*/
delRouteAndARP();             /*Deletes all the routes and ARP entries*/
readRoutes (DATAFILE2);       /*It reads the default routes*/
initsockets();                /*It initialites all the sockets*/
mynode.FA.COA= 0; 
mynode.FA.seq_num=0;
mynode.FA.Lifetime=0;
mynode.FA.TimeBetweenAdverts=0;
mynode.FA.flags = 0;
mynode.FaOrHa=0; /*I am (was) connected to nowhere*/
mynode.status = DOWN; /*I suppose when I switch the computer on, it has to register*/
mynode.identification[0] = (unsigned long) 0;
mynode.SPI = 1122; /*SPI*/
devARPrequest (2, INTERFACE);
startTimers();
seed=0;sAgentSol();
DEBUG (3) printf ("Init all procedure finished\n");
}

/********************************************************************/

void disconnection()
{
  
  /*When the mobile node is disconnected to the network (because it does not
    receive any agent adversiments), this function is called. In that case, 
    mobile node assumes it is at home but it sends agent advertisements*/

  /*clean up ForeignA*/
  DEBUG (1) printf ("Disconnection with the FA/HA (Agents)\n");
  mynode.FA.seq_num=0;
  mynode.FA.Lifetime=0;
  mynode.FA.TimeBetweenAdverts=0;
  mynode.FA.flags = 0;
  if (mynode.status == WAITING) cancelTimer (mynode.timeout);
  else if ((mynode.status == UP) && (mynode.FaOrHa == 2)) 
		{
			cancelTimer (mynode.regTimeout);
			#ifdef ARP_REPLIES
				cancelTimer (mynode.ARPRepTimeout);
			#endif
		}
  mynode.status = DOWN;
  mynode.FaOrHa = 0;
  mynode.FA.COA = (unsigned long)0;
  delRouteAndARP();
  readRoutes(DATAFILE2);
  seed=0;sAgentSol();
}

/********************************************************************/

void Registration (int lifet, unsigned long Address)

{

  /*This procedure is used to send registration requests to the home agent*/

struct sockaddr whereto;
struct sockaddr_in *to;
struct MIPRegReq *RegisReq;
unsigned long temp;
char packet[1024];
int i,value,sent;
struct timeval *tv;

bzero((char *)&whereto, sizeof (struct sockaddr));
to = (struct sockaddr_in *)&whereto;
to->sin_family = AF_INET;
to->sin_addr.s_addr = Address ;
to->sin_port = htons(AGENT_LISTEN_PORT);
RegisReq = (struct MIPRegReq *)packet;

RegisReq->type = 1; /*Registration Request value*/
RegisReq->flags = 0; /*It should be different*/
RegisReq->lifetime = htons(lifet);
RegisReq->Address = mynode.Address;
RegisReq->homeAgent = mynode.HA.Addr;
RegisReq->careOfAddress = Address; /*But it's at home ... COA means ... what?*/
/*gettimeofday (tv, NULL);*/
RegisReq->identification[0] = htonl((unsigned long)newId());
mynode.identification[1] = RegisReq->identification[1] = htonl((unsigned long)newId());
DEBUG (1)
 {
   printf ("	Identification0 %lu\t", RegisReq->identification[0]);
   printf ("Identification1 %u\n",RegisReq->identification[1]);
 }
value = sizeof (struct MIPRegReq);

/*Mobile-Home authentication*/ 

*(packet+value++) = 32;
*(packet+value++) = 4 + LENGTHMD5; /*Data length of mobile-home autentication extension*/
temp = htonl(mynode.SPI);
bcopy (&temp,(packet+value), 4); /*SPI*/
appendauth(packet,value,mynode.MD5key,mynode.MD5keylength);  
value += (4 + LENGTHMD5);


/*Mobile-Foreign Authentication*/

if (authenAssociation (mynode.FA.COA, mynode.MNFA) && (mynode.FaOrHa == 2))
  /*If a security association exists between MN and FA and MN is not at home then
    Mobile-Foreign Authentication extension must be appended*/
  {
    /*Create extension */
    struct authentication *foreignKey;

    foreignKey = authenAssociation (mynode.FA.COA, mynode.MNFA);

    *(packet + value++) = 33;
    *(packet + value++) = 4 + LENGTHMD5;
    temp = htonl(foreignKey->SPI); /*SPI*/
    bcopy (&temp, (packet + value), 4);
    appendauth (packet, value, foreignKey->MD5key, foreignKey->MD5keylength);
    value += 4 + LENGTHMD5;
    DEBUG (1) printf ("Mobile-Foreign authentication has been placed\n");
  }


/*I have to send an UDP packet*/
sent = sendto (SockAgentInOut, packet, value ,0, &whereto, sizeof (struct sockaddr)); 
if ((sent < 0) || (sent != value)) perror ("Registration() sendto error");
else DEBUG(1) printf ("I've sent a registration packet to my HA/FA\n");
}

/********************************************************************/

void gettingMACFA (unsigned long pCOA)
{

  /*This procedure is used to obtain foreign agent's MAC address from the
    agent advertisement packet*/

struct ext
{
   unsigned char type;
   unsigned char length;
   unsigned short seq_num;
   unsigned short lifetime;
   unsigned short flags;
   unsigned long COA;
};

struct ethhdr *h;
struct sockaddr_in fromOutside;
struct ext *service;
struct icmphdr *icmp;
int packet,sock, value, fromlength, lengthIp;
char meto[1024];
struct iphdr *ip;
unsigned char *num_addr, *addr_entry_size;

bzero ((char *)&fromOutside, sizeof (struct sockaddr_in));
sendSol();

sock = socket (AF_INET, SOCK_PACKET, htons (ETH_P_ALL));
if (sock <0 )
{
   perror ("Ethernet socket.");
   exit (-1);
}
fromlength= sizeof (fromOutside);
packet = 600;
for (;;)
 {
   value = recvfrom (sock, (char *)meto, packet, 0, (struct sockaddr *)&fromOutside, &fromlength);
   ip = (struct iphdr *)(meto +sizeof (struct ethhdr));
   icmp = (struct icmphdr *)(meto + sizeof (struct ethhdr) + sizeof (struct iphdr));
   if (value < sizeof (struct icmphdr) + sizeof (struct iphdr) + sizeof (struct ethhdr))
          continue;
   if ((icmp->type == 9) && (ip->protocol == 1) && ((icmp->code == 0) ||
   					(icmp->code == 16)))
    {
	h = (struct ethhdr *)meto;
	ip = (struct iphdr *)(meto + sizeof (struct ethhdr) );
	lengthIp = ip->ihl << 2; /*It gets in lengthIp the number of octects*/
	icmp = (struct icmphdr *)(meto + lengthIp + sizeof (struct ethhdr));
	if (value < (sizeof(struct iphdr) + sizeof(struct icmphdr)+
	             sizeof(struct ethhdr)))
	   continue;
	num_addr = (unsigned char *)(meto + lengthIp + 4 + sizeof (struct ethhdr));
	addr_entry_size = (unsigned char *)(meto + lengthIp + 5 + sizeof (struct ethhdr));
	service = (struct ext *)(meto + lengthIp + 8 + (*num_addr * *addr_entry_size * 4) +
	           sizeof (struct ethhdr));
    	if (service->type != 16) 
    	    continue;
    	if (service->length == 6) service->COA = fromOutside.sin_addr.s_addr;
    	if (service->COA == pCOA)
    	    {
		if (fullyc) devARPrequest (2, INTERFACE);
    	        createARP (service->COA,h->h_source); 
		printf ("Putting in the ARP cache %s\n", inet_ntoa(service->COA));
		#ifdef ARP_REPLIES
		      bcopy (h->h_source, mynode.FA.hwAddress, 6);
		#endif
		DEBUG (3) printf ("MAC address inserted in ARP cache\n");
		ARPreply (service->COA,h->h_source, mynode.Address, mynode.MAC);
		ARPreply (service->COA,h->h_source, mynode.Address, mynode.MAC);
		mynode.FA.Lifetime = ntohs (service->lifetime);
		close (sock);
    		return;
 	    }   
    }
 }
close (sock);
}


/********************************************************************/

void CheckAgentAdvert(char *packet,int length,struct sockaddr_in from)
{

  /*This procedure is used to check that the agent advertisment is right and
    it acts depending on it*/

struct icmphdr *icmp;
struct iphdr   *ip;
int lengthIp;
int lifeTimeAdvert,*temp;
unsigned char *num_addr, *addr_entry_size;

struct extension
{
   unsigned char type;
   unsigned char length;
   unsigned short seq_num;
   unsigned short lifetime;
   unsigned short flags;
   unsigned long COA;
};

struct extension *service;
struct advert *adv;

ip = (struct iphdr *)packet;
lengthIp = ip->ihl << 2; /*It gets in lengthIp the number of octects*/
icmp = (struct icmphdr *)(packet + lengthIp);
num_addr = (unsigned char *)(packet + lengthIp + 4);
addr_entry_size = (unsigned char *)(packet + lengthIp + 5);
service = (struct extension *)(packet + lengthIp + 8 + (*num_addr * *addr_entry_size * 4));

if (ip->protocol !=1 || icmp->type !=9 || icmp->code !=0)
    {
      DEBUG (4) printf ("It's not a icmp packet (%d) or it is not an agent advert (%d).\n",icmp->type, icmp->code);   
      return; /*It is not a valid ICMP packet*/
    }

if (!checkChecksumICMPAdvert(icmp, length-lengthIp))
{
  DEBUG(1) printf ("Checksum incorrect.Discarding packet\n");
  return;
}

if (length < lengthIp + 10+ (*num_addr * *addr_entry_size * 4)+ service->length)
     {
      DEBUG (1) printf ("Packet too short!\n");
      return; /*Invalid mobile service Extension. Packet too short. */
     }

if ((ntohs(service->flags) & FFLAG) && !(service->length-6))
   
  {
   DEBUG (1) printf ("It's a FA but it does not carry any COA\n"); 
   DEBUG (1) printf ("Using %d as COA ... \n", inet_ntoa(from.sin_addr.s_addr));
   service->COA = from.sin_addr.s_addr;
  }
/*At this point, the Mobile Service Extension packet is ok!*/

if (ntohs(service->flags) & BFLAG) /*Agent too busy*/

  {
   DEBUG (1) printf ("Agent busy .... exiting\n");
   return;
  }
DEBUG (3)
  {
    printf ("Home agent flag %s, Foreign A. flag: %s\n",ntohs(service->flags) & HFLAG ? "yes" : "no",
                ntohs(service->flags) & FFLAG ? "yes" : "no");
    printf ("Home agent address %s ",inet_ntoa(mynode.HA.Addr));
    printf ("Foreign Agent addr: %s\n", inet_ntoa(mynode.FA.COA));
    printf ("Home NetMask %s\n",inet_ntoa(mynode.HA.NetMask));
  }
adv = (struct advert *) (packet + lengthIp + 4);
lifeTimeAdvert = ntohs (adv->Lifetime);

#ifdef SEVERAL_AGENTS_OPTION

if (((mynode.FA.COA == service->COA) && (mynode.FaOrHa)) ||
   (((mynode.HA.Addr == service->COA) || (mynode.HA.Addr==from.sin_addr.s_addr))
    && (mynode.FaOrHa == 1)))
      {
          cancelTimer (mynode.timerId);
          mynode.timerId = addTimer (lifeTimeAdvert,
          disconnection, NULL,NULL);
      }
else if (!mynode.FaOrHa) 
  mynode.timerId = addTimer (lifeTimeAdvert, disconnection, NULL,NULL);

#endif

#ifndef SEVERAL_AGENTS_OPTION

if (mynode.FaOrHa)
         cancelTimer (mynode.timerId);
DEBUG (5) printf ("The lifetimeAdvert is: %d\n", lifeTimeAdvert);
mynode.timerId = addTimer (lifeTimeAdvert, disconnection, NULL,NULL);

#endif

if ((ntohs(service->flags) & HFLAG) && ((service->COA == mynode.HA.Addr) || (from.sin_addr.s_addr ==
							 mynode.HA.Addr)))
 {
   DEBUG (4) 
     {
      printf ("Home Agent advertisement. Address: %s\n",inet_ntoa(service->COA));
      printf ("I am at home. The agent address is my home address!\n"); 
     } 
   if (idSol!=-1) 
    {
   DEBUG (0) 
      printf ("I was sending solitations. I've got an agent advert, so I'll stop\n");
     
     cancelTimer (idSol);
     idSol=-1;
    }
   if ((mynode.FaOrHa == 2) || ((mynode.FaOrHa == 0) && (mynode.status == DOWN))) 
     {
       DEBUG (1) 
        {
          if (mynode.FaOrHa) printf ("I was away from home, but now, I've come back!\n");
          else printf ("I was disconnected. Now I'm at home and I dont know where I was b4 (maybe here)\n");
        }
	/*COA must be stored in some place, because with the Reg. Rep. we have not information about COA*/
       tempCOA = (unsigned long) 0;
       if (mynode.status == UP)
	{
                 cancelTimer (mynode.regTimeout);
		 #ifdef ARP_REPLIES
			cancelTimer (mynode.ARPRepTimeout);
		 #endif
	}
       if (mynode.status == WAITING) cancelTimer (mynode.timeout);
       if (fullyc) 
	{
		devARPrequest (2, INTERFACE);
	}
       if (mynode.FaOrHa)
         {
	   delRouteAndARP();
           readRoutes(DATAFILE2); 
         }
       mynode.FA.COA = 0;
       Registration(0, mynode.HA.Addr);
       DEBUG (4) printf("A timer is created in order to wait for a registration reply\n");
       mynode.timeout = addTimer (WAITINGANSWER, changeState, NULL,NULL);
       mynode.status = WAITING; /*It means pending*/
     /*Registration with HA, lifetime=0, mynode.HA.Addr */ 
       
     }
   else
     {
      /*I was at home b4!*/
      mynode.HA.Lifetime=ntohs(service->lifetime);    
      DEBUG (3) 
        printf ("I was at home b4 and I've got another packet from my home agent\n");
     } 
  mynode.FaOrHa = 1; /*we are at home (1) */
  return; 
 }
if (ntohs(service->flags) & FFLAG)
 {
   if ((service->COA == mynode.FA.COA) && (mynode.status == UP))

    {
      /*I was here, and I'm still here*/
       if ((mynode.FA.seq_num > ntohs(service->seq_num)) && (ntohs(service->seq_num) <=255)) 
       {
         DEBUG (-1) 
           {
             printf ("Warning ... FA was down!!! I have to register again\n");
             printf ("Problems with secuence numbers and rollover!\n");
           }
         tempCOA = mynode.FA.COA;
         mynode.FA.seq_num = ntohs(service->seq_num);
         cancelTimer (mynode.regTimeout);
         Registration (mynode.FA.Lifetime, mynode.FA.COA);
         mynode.timeout = addTimer (WAITINGANSWER, changeState, NULL, NULL);
         mynode.status = WAITING;
         return; /*sure*/ 
       }
     
      DEBUG (1) printf ("----\n");
      DEBUG (3) 
      {
        printf ("I'm on a foreign net and I've got another packet from \n");
        printf ("the same FA where I was registered. So, I have to update some values.\n");
      }
      
      /*I was here b4. Therefore I just have to update some fields*/
      mynode.FA.Lifetime=ntohs(service->lifetime);    
      mynode.FA.seq_num=ntohs(service->seq_num);
     }
   else if (!(mynode.FA.COA) && ((service->COA & mynode.HA.NetMask) == 
   		(mynode.HA.Addr & mynode.HA.NetMask)))
          {
  	    DEBUG (2) printf ("I dont need a FA in my home network!\n");
            return; /*I dont need a FA in my home network!*/
          }
   	     else
   	       {
   	          DEBUG (1) printf ("I HAVE CHANGED THE NETWORK\n");
   	          /*I think it is time to register!*/
		 if ((mynode.status == UP) && (mynode.FaOrHa == 2)) 
			{
				cancelTimer(mynode.regTimeout);
				#ifdef ARP_REPLIES
					cancelTimer (mynode.ARPRepTimeout);
				#endif
			}
                 if ((mynode.status == WAITING) && (mynode.FA.COA !=service->COA)) 
		 {
		       cancelTimer (mynode.timeout);
                       mynode.status = DOWN;
                 }
		 if (idSol!=-1) 
    		  {
 		    DEBUG (1) printf ("I give up sending Solicitation Packets\n"); 
     		    cancelTimer (idSol);
     		    idSol=-1;
    		  }
		    
		    mynode.FA.COA = service->COA;
     		    mynode.FA.Lifetime=ntohs(service->lifetime);
     		    mynode.FA.flags=ntohs(service->flags);
     		    mynode.FA.seq_num=ntohs(service->seq_num);
	            mynode.FaOrHa = 2;
	            DEBUG (4) 
	             {
	               printf ("\nNew Network: %s\n",inet_ntoa(service->COA));
	               printf ("Lifetime %d\n",mynode.FA.Lifetime);
	               printf ("flags %u Seq num %u\n",mynode.FA.flags,mynode.FA.seq_num);
	              
	             }
		    tempCOA = mynode.FA.COA;
		    if (mynode.status != WAITING)
                     {
			char hwaddr[7];
			
			delRouteAndARP();
                        routeEntryDev(mynode.FA.COA,INTERFACE,inet_addr("255.255.255.255"));
                        routeEntry(1, (unsigned long) 0, mynode.FA.COA,(unsigned long)0);
			routeEntryDev((unsigned long)inet_addr("255.255.255.0"), INTERFACE,inet_addr ("255.255.255.255"));
			/*I have to check if service->COA already exists*/
			if (!requestARP(service->COA, hwaddr))
			     gettingMACFA (service->COA);	
                        Registration (mynode.FA.Lifetime, mynode.FA.COA);
			mynode.timeout = addTimer(WAITINGANSWER, changeState, NULL, NULL);
                     }
       		    mynode.status = WAITING; /*pending*/
                    return;
	       }
 }  
}

/*********************************************************************/

void changeState ()
{

  /*This function is called when the mobile node has issued a registration 
    request but it did not get any answer from the foreign agent*/

  DEBUG (1) printf ("TIME OUT:
       I HAVENT GOT ANY REGISTRATION REPLY\n");
  mynode.status = DOWN; /*I havent got a Registration Reply*/
  mynode.FA.COA = (unsigned long)0;
}

/*********************************************************************/

void registrationAgain()
{

  /*This function is called when the mobile node is registered with its home
    agent (and foreign agent) and it has to register again, because the
    lifetime is near to expire*/

Registration (mynode.FA.Lifetime, mynode.FA.COA);
mynode.timeout = addTimer (WAITINGANSWER,changeState,NULL,NULL);
mynode.status = WAITING; /*pending*/

}

/*********************************************************************/

void registrationReply(char *packet, int length)
{

  /*This function is called when the mobile node has obtained a registration
    reply from its home agent. it checks all the values and it acts depending
    on those values*/

struct MIPRegRep *Reply;
int detectederror = 0;
int extensions;

Reply = (struct MIPRegRep *)packet;

if ((mynode.status != WAITING) || (Reply->type != 3)) return;
if (Reply->Address != mynode.Address)
     return;  /*This registration response is not for me!(How did I get it????)*/
if (((!Reply->code) || (Reply->code == 1) || (Reply->code>127))
     && (length > sizeof (struct MIPRegRep)))
 {
   extensions = sizeof (struct MIPRegRep);

   while ((*(packet +extensions) !=32) && (extensions < length))
       extensions += 2 + *(packet + extensions + 1);
   if (extensions >= length) detectederror = 1;
   if  (*(packet + extensions) == 32)
    {
      unsigned char MD5keyfound[KEYLENGTH];
      unsigned long *SPI;

      bcopy ((packet + extensions + 6), MD5keyfound, LENGTHMD5);
      SPI = (unsigned long *) (packet + extensions + 2);
      if ((*SPI >= 0) && (*SPI <=255))
        {
	   DEBUG (1) printf ("Registration Reply (HA-MN): SPI error\n");
	   return;
	}   
      appendauth (packet, extensions + 2, mynode.MD5key, mynode.MD5keylength);
      if (bcmp(MD5keyfound, packet + extensions + 6, LENGTHMD5))
       {
         DEBUG (0) printf ("Error: MN-HA authentication not valid\n");
         return;
       }
     }
 }
else if (!((Reply->code >2) && (Reply->code <128))) 
{
     return;
}

if (authenAssociation(mynode.FA.COA, mynode.MNFA))
    {
      extensions = sizeof (struct MIPRegRep);

      while ((*(packet + extensions) != 33) && (extensions < length))
         extensions += 2 + *(packet + extensions + 1); 
      if (extensions >= length) detectederror = 1;
      if (*(packet + extensions) == 33)
       {
         unsigned char MD5keyfound[KEYLENGTH];
         struct authentication *position;
	 unsigned long *SPI;
     
         position = authenAssociation (mynode.FA.COA, mynode.MNFA);
         bcopy ((packet + extensions + 6), MD5keyfound, LENGTHMD5);
	 SPI = (unsigned long *)(packet + extensions + 2);
	 if ((*SPI >= 0) && (*SPI <= 255))
	  {
		DEBUG (1) printf ("Registration Request (FA-MN): SPI mismatch\n");
		return;
	  }

         appendauth (packet, extensions + 2, position->MD5key, position->MD5keylength); 

         if (bcmp(MD5keyfound, packet + extensions + 6, LENGTHMD5))
          {
            DEBUG(0) printf ("Error: MN-FA authentication not valid\n");
            return;
          }
        }
    }

if ((Reply->code !=0) && (Reply->code !=1))
 {
   switch (Reply->code)
    {
      case 67: printf ("MN-FA authenticator was required\n"); break;
      case 131: printf ("MN-HA authenticator failed\n");break;
      case 70: printf ("Poorly formed request\n");break;
      case 69: printf ("Lifetime too long\n");break;
      default: DEBUG (1) printf("Registration reply not valid. Value %d\n",
      				 Reply->code);
    }
   detectederror = 1;
 }

mynode.identification[0] = Reply->identification[0];

DEBUG (1)
  printf ("Identification0 %u Ident1 %u\n",Reply->identification[0],
  	   Reply->identification[1]);

/*
if (Reply->identification[1] != mynode.identification[1])
  {
       printf ("Identifications do not match!\n");
       detectederror = 1;
  }
*/
/*Check MN-HA authentication*/

if (detectederror)
  {
    cancelTimer (mynode.timeout);
    if (Reply->code == 69)
     {
       DEBUG (2) printf ("New lifetime for a new Registration %d\n", ntohs(Reply->lifetime));
       mynode.FA.Lifetime = ntohs(Reply->lifetime);
       registrationAgain();
       mynode.FaOrHa = 2;
       return;
     }
    
    mynode.FaOrHa = 2;
    mynode.status = DOWN;
    return;
  }

if (mynode.status == WAITING)
 {
   if (mynode.FA.Lifetime <= ntohs(Reply->lifetime)) mynode.HA.Lifetime = mynode.FA.Lifetime;
   else 
       mynode.HA.Lifetime = mynode.FA.Lifetime = ntohs(Reply->lifetime);
   mynode.FA.COA = tempCOA;
   mynode.status = UP; 
   cancelTimer(mynode.timeout);
   DEBUG (1) printf ("I've got a packet from my FA/HA. I'm registered\n");
   if (Reply->lifetime) /*If I'm at home I dont have to register anymore*/
    {
      mynode.regTimeout = addTimer (mynode.HA.Lifetime/4, registrationAgain, NULL, NULL);
      #ifdef ARP_REPLIES
	mynode.ARPRepTimeout = addTimer (300 /*5 mins*/, ARP_send_background, NULL, NULL);
      #endif
      routeEntry (1, (unsigned long)0 /*It means default*/, mynode.FA.COA, (unsigned long) 0);
      /*A route has been set up. Target: MN (default) to FA*/
      if (fullyc) devARPrequest (1, INTERFACE);/*It does not answer to ARP requests*/
      mynode.FaOrHa = 2;
    } 
   else 
     {
       DEBUG (3)
         {
            printf ("Issuing a gratuituos ARP\n");
            printf ("Address %s. MAC address %x:%x:%x:%x:%x:%x\n", inet_ntoa(mynode.Address), mynode.MAC[0],
		     mynode.MAC[1], mynode.MAC[2], mynode.MAC[3], mynode.MAC[4], mynode.MAC[5]);
	 }
	 GratuitousARP (mynode.Address, mynode.MAC);
	 addTimer (2, gratARPbackground, (void *)&mynode.Address, (void *)mynode.MAC);
         addTimer (4, gratARPbackground, (void *)&mynode.Address, (void *)mynode.MAC);
	 mynode.FaOrHa = 1;       
     } 
   DEBUG (5) printf ("My COA is %s\n",inet_ntoa(mynode.FA.COA));
   return;
 }
}

/********************************************************************/
/********************************************************************/
/********************************************************************/


main(int argc, char *argv[])
{
      fd_set fdvec;
      struct timeval tv;
      int selcode;
      int packlen = MAX_PACKET;

      switch (argc)
	{
		case 3 :  
			debugLevel = atoi (argv[1]);
          		fullyc = atoi (argv[2]);
			break;
		case 2 :
			debugLevel = atoi (argv[1]);
			break;
		case 1: 
			debugLevel = -1;
			break;
		default:
			printf ("%s [DebugLevel] [0(=demo)|1(=fullycompiant)]");
			exit (1);
	}

      DEBUG (1) printf ("Mobile Node running .... DebugLevel %d Mode: %s\n", debugLevel, fullyc ? "Fully compliant" : "demo ");
      initall();           /*Initialize*/
     for (;;)
     {
      FD_ZERO(&fdvec); 
      FD_SET(SockAgentInOut, &fdvec); 
      FD_SET(SockAdverts, &fdvec); 
      tv.tv_sec = 0; tv.tv_usec = 10;
      while((selcode = select(FD_SETSIZE, &fdvec, NULL, NULL, (struct timeval *)0)) == -1)  
       {

	 if(errno == EINTR ||  errno == ERESTART);
	 else  
	 perror("select() failed in main loop.\n");
	  
       }
      if (FD_ISSET(SockAdverts, &fdvec))
       {
         struct sockaddr_in from;
	 int length, fromlen;

	 fromlen = sizeof(from);
	 if ((length = recvfrom(SockAdverts, (char *)packet, packlen, 0,
	      (struct sockaddr *)&from, &fromlen)) < 0) 
	  {
	    if (errno == EINTR) continue;
	    DEBUG (3) perror("recvfrom error");		
	  }


	 DEBUG (5) printf ("I've got a packet(SockAdverts)\n");
	 CheckAgentAdvert(packet, length, from);    
       }
     if (FD_ISSET(SockAgentInOut, &fdvec))
      {
        struct sockaddr_in from;
        int length, fromlen;
  
        fromlen = sizeof (from);
        if ((length = recvfrom(SockAgentInOut, (char *)packet, packlen, 0,
	       (struct sockaddr *)&from, &fromlen))  < 0)
          {
             if (errno ==EINTR) continue;
	     DEBUG (3) perror ("recvfrom error");
          }
        registrationReply (packet, length);
      }

 }
}
/********************************************************************/

int checkChecksumICMPAdvert (struct icmphdr *icmp, int length)
{
unsigned short swapvar;

swapvar=icmp->checksum;
icmp->checksum = 0;
icmp->checksum = in_cksum (icmp, length);
if (swapvar != icmp->checksum) return 0;
return 1;

}







