/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/

/*This module is used in myagent.c*/

void routeEntry (int, unsigned long, unsigned long, unsigned long);
void delARPentry (unsigned long, int proxy);

extern struct HAvalues HomeA;

/********************************************************************/

void changeTunl (int index, unsigned long Address)
{

  /*This procedure is used to change the end point of a tunnel device. It is
    useful when a mobile node has changed from a foreign network to another
    foreign network*/

   int ioctlcall;
   struct sockaddr_in *sock;
   char device[10];
   struct ifreq request;

   bzero ((char *)&request, sizeof (struct ifreq));
   sprintf (device, "tunl%c",index+'0');
   strncpy (request.ifr_name, device, strlen(device));

   sock = (struct sockaddr_in *) &request.ifr_dstaddr;
   sock->sin_family = AF_INET;
   sock->sin_addr.s_addr = Address;

   ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
   if (ioctl(ioctlcall, SIOCSIFDSTADDR, (char *)&request) == -1)
     perror ("Change tunnel error");

   close (ioctlcall);
}

/********************************************************************/

void addIFtunl (int tunnel, unsigned long Address)
{

  /*This procedure is called when the home agent creates a tunnel interface.
    The value specefied in Address, is the endpoint of the tunnel*/

  struct sockaddr_in *sock;
  struct ifreq request;
  int ioctlcall,count;
  char device[10];
  
  bzero ((char *)&request, sizeof (struct ifreq));
  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

  count = sprintf (device, "tunl%c", tunnel+'0');
  device[count] = '\0';
  
  bcopy (device, request.ifr_name, 6);
  sock = (struct sockaddr_in *)&request.ifr_addr;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = HomeA.Addr;
  if (ioctl(ioctlcall, SIOCSIFADDR, (char *)&request) == -1)
    perror ("Error creating a tunnel");

  sock = (struct sockaddr_in *)&request.ifr_dstaddr;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;
  
  if (ioctl(ioctlcall, SIOCSIFDSTADDR, (char *)&request) == -1)
    perror ("Error creating a tunnel (2)");

  request.ifr_flags = (IFF_NOARP | IFF_UP | IFF_POINTOPOINT);
  if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) == -1)
    perror ("Error creating the tunnel (3)");
 
  close (ioctlcall);

}

/********************************************************************/

void routeEntryDev (unsigned long Address,char *device, unsigned long Mask)
{

  /*This function is called to create an entry in the routing table. The origin
    of the entry is Address and packets which are addressed to that address are
    sent to the interface specified in device*/

  struct sockaddr_in *sock, *sockmask;
  struct rtentry request;
  int ioctlcall;
  int count;  

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
  bzero ((char *)&request, sizeof (struct rtentry));

  sock = (struct sockaddr_in *) &request.rt_dst;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;

  sockmask = (struct sockaddr_in *) &request.rt_genmask;
  sockmask->sin_family = AF_INET;
  sockmask->sin_addr.s_addr = Mask;

  request.rt_flags = (RTF_UP | RTF_HOST);
  request.rt_dev = device;
  if (ioctl(ioctlcall, SIOCADDRT, (char *)&request) == -1)
    perror ("Route entry can not be added");
  close (ioctlcall);
}

/********************************************************************/

void delIFtunl (int tunnel)
{

  /*This procedure removes a tunnel device*/

   char device[10];
   int ioctlcall;
   struct ifreq request;

   bzero ((char *)&request, sizeof (struct ifreq));
   ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

   device[sprintf (device, "tunl%c",tunnel + '0')] = '\0';
   bcopy (device, request.ifr_name, 6);
   
   if (ioctl(ioctlcall, SIOCSIFADDR, (char *)&request) <0)
     perror ("tunnel can not be deleted");
   
   request.ifr_flags &= ~IFF_UP;
   if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) == -1)
     perror ("tunnel can not be deleted (2)");
   close (ioctlcall);

}

/********************************************************************/

void proxyARP (unsigned long Address, char *hwAddress)
{

  /*It forces the creation of a proxy entry in the home agent as permanent.
    the IP address is specified in Address and the MAC address is specified in
    hwAddress*/

  int ioctlcall;
  struct sockaddr_in *sock;
  struct arpreq request;

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
 
  bzero ((char *)&request, sizeof (struct arpreq));
  sock = (struct sockaddr_in *) &request.arp_pa;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;

  sock = (struct sockaddr_in *)&request.arp_ha;
  sock->sin_family = ARPHRD_ETHER;
  bcopy (hwAddress, request.arp_ha.sa_data, 6);
  request.arp_flags = (ATF_PERM | ATF_PUBL | ATF_COM);
  if (ioctl(ioctlcall, SIOCSARP, (char *)&request) < 0)
      perror ("Proxy ARP has not been done");
  close (ioctlcall);
}

/********************************************************************/

void routeEntry (int type, unsigned long target, unsigned long dest,
		 unsigned long Mask)
{

  /*This function is created to handle routing entries. if the value of the
    type file is:
    1) It creates a route
    2) It deletes a route
    3) it creates a route to a specific interface.
   */

int ioctlcall;
struct rtentry rtent;
struct arpreq ar;
struct sockaddr_in *sock, *socket2, *sockMask;

bzero ((char *)&rtent, sizeof (struct rtentry));

sock = (struct sockaddr_in *)&rtent.rt_dst;
sock->sin_family = AF_INET;
sock->sin_addr.s_addr = target;

socket2 = (struct sockaddr_in *)&rtent.rt_gateway;
socket2->sin_family = AF_INET;
socket2->sin_addr.s_addr = dest;

sockMask = (struct sockaddr_in *)&rtent.rt_genmask;
sockMask->sin_family = AF_INET;
sockMask->sin_addr.s_addr = Mask;

ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

switch (type)
{
 case 1:
   
     rtent.rt_flags = RTF_UP |  RTF_GATEWAY;

     if (ioctl(ioctlcall, SIOCADDRT, (char *)&rtent) == -1)
        perror ("SIOCADDRT failed");
     break;
  

 case 2:
  
     if (ioctl(ioctlcall, SIOCDELRT, (char *) &rtent) < 0 && errno!= EFAULT)
            DEBUG (3) perror ("Delete route error\n");
     break; 

 case 3:
     rtent.rt_flags = RTF_UP | RTF_HOST;
     bcopy (interface, rtent.rt_dev,strlen(interface));
     if (ioctl(ioctlcall, SIOCADDRT, (char *)&rtent) == -1)
         perror ("I cannot create a route");
     break;
}
close (ioctlcall);
}

/*********************************************************************/

/*********************************************************************/

void delRouteAndARP()
{
/*This function is not used by myagent ... must be removed*/

   FILE *router;
   char string[300],device[20], ipaddress[20];
   unsigned long dest, gateway, mask;
   unsigned int a1, a2, a3, a4;

/*This function is used to remove all the entries in the routing table 
(/proc/net/route) and all
the entries in the ARP cache (/proc/net/arp)*/

   router = fopen("/proc/net/route","r");
        
   fgets (string, 255, router);
   while (fgets(string, 255, router) != NULL)
     {
        if (sscanf(string, "%s %lx %lx %u %u %u %u %lx ", device, &dest, &gateway,
		&a1, &a2, &a3, &a4, &mask) != 8) continue;
        if (strcmp(device,"lo"))
            routeEntry (2, dest, gateway, mask);
     }
   fclose (router);

   router = fopen ("/proc/net/arp","r");
   fgets(string, 255, router);
   while (fgets(string, 255, router) != NULL)
     {
        if (sscanf (string, "%s ",ipaddress) != 1) continue;
        DEBUG (3) printf ("Address to delete from ARP cache: %s\n",device);
        if ((dest = inet_addr(ipaddress)) != -1) 
	       delARPentry(dest, 1);	
     }
   fclose (router);
}

/********************************************************************************/

void delARPentry (unsigned long ipaddress, int proxy)
{

/*Given one IP address, this function deletes that IP address from the ARP cache
if proxy == 0 deletes a normal entry if 1 then deletes a proxy arp entry
*/


   struct arpreq request;
   struct sockaddr_in *sock;
   int ioctlcall;
   
   ioctlcall = socket (AF_INET,  SOCK_DGRAM, 0);

   bzero ((char *)&request, sizeof(struct arpreq));
   sock = (struct sockaddr_in *) &request.arp_pa;
   sock->sin_family = AF_INET;
   sock->sin_addr.s_addr = ipaddress;
   if (proxy) request.arp_flags = ATF_PUBL;
   
   DEBUG (3) printf ("Address: %s\n",inet_ntoa(ipaddress));
   if (ioctl(ioctlcall, SIOCDARP, (char *) &request) == -1)
      perror ("It cannot delete an entry in ARP cache");
   else DEBUG (3)
            printf ("ARP Entry removed ...\n");
   close (ioctlcall);
}

/********************************************************************/






