/*
  This file may be copied, modified and distributed only in accordance with
  the terms of the limited licence contained in the accompanying file <licence.txt>

*/

/*This module is used in mymnode.c*/

void routeEntry (int, unsigned long, unsigned long, unsigned long);
void delARPentry (unsigned long, int);
void routeEntryDev (unsigned long, char *,unsigned long);

/********************************************************************/

void readRoutes (char *name)
{

  /*This procedure reads the default routes specified in the routeentry.conf
    file.*/

  FILE *fp;
  char device[20], target[30], gateway[30], mask[30];
  char i, number;

  fp = fopen (name,"r");

  if (fp == NULL)
   {
	printf ("File %s missing!\n", name);
	exit (-1);
   }
  
  number = readtoken (fp, device, 0);

  for (i=0; i< number; i++)
   {
     readtoken (fp, device, 1);
     readtoken (fp, target, 1);
     readtoken (fp, gateway, 1);
     readtoken (fp, mask, 1);
 
     routeEntry (1, inet_addr(target), inet_addr(gateway), inet_addr(mask));
     DEBUG (5) printf ("Device %s, target %s, gateway %s\n", device
			, target, gateway);
   }
  routeEntryDev (inet_addr("224.0.0.1"), "eth0", inet_addr("255.255.255.255"));
  fclose (fp);
}

/********************************************************************/

void readValuesFromFile (char *name)
{
  /*This procedure reads the values defined in the configuration file and
    it fills the appropiate variables*/

  FILE *fp;
  char line[300],device[20],target[16],gateway[16];
  int i,number;
  struct authentication *create;
 
  fp = fopen (name,"r");

  if (fp == NULL)
   {
        printf ("File %s missing!\n", name);
        exit (-1);
   }

  readtoken (fp, target, 1);
  readtoken (fp, gateway, 1);
  mynode.Address = inet_addr(target);
  mynode.HA.Addr = inet_addr(gateway); 

  DEBUG (3) 
    {
	printf ("My address is %s ",inet_ntoa (mynode.Address));
        printf ("and my HA is %s\n",inet_ntoa(mynode.HA.Addr));
    }
  readtoken (fp, interface, 1);
  readtoken (fp, line, 1);
  sscanf(line, "%x:%x:%x:%x:%x:%x\n",&mynode.MAC[0],&mynode.MAC[1], &mynode.MAC[2], &mynode.MAC[3],
		  &mynode.MAC[4], &mynode.MAC[5]);

  readtoken (fp, line, 1);
  mynode.HA.NetMask = inet_addr (line);
  DEBUG (4) printf ("Reading MN-HA assotiation\n");
  mynode.MD5keylength = readtoken (fp, line, 0); 
  readtoken (fp, line, 1);
  strcpy (mynode.MD5key, line);

  DEBUG (4) printf ("Key size %d. Value: %s\n",mynode.MD5keylength, mynode.MD5key);
  DEBUG (4) printf ("Reading MN-FA assotiations\n");

  number = readtoken (fp, line, 0);
  mynode.MNFA = NULL;
  for (i=0; i< number; i++)
   {
      create = (struct authentication *)malloc(sizeof (struct authentication));
      readtoken (fp, line, 1);
      create->Address = inet_addr (line);
      create->SPI = readtoken (fp, line, 0);   /*SPI*/
      create->MD5keylength = readtoken (fp, line, 0);
      readtoken (fp, line, 1);
      bcopy (line, create->MD5key, create->MD5keylength);
      create->next = mynode.MNFA;
      mynode.MNFA = create;
      DEBUG (4) printf ("Address %s, keylength %d key %s\n", inet_ntoa(create->Address),
			 create->MD5keylength, create->MD5key);
   } 
  fclose (fp);
}

/********************************************************************/

void routeEntryDev(unsigned long target, char *device, unsigned long Mask)
{

  /*This function is called to create an entry in the routing table. The origin
    of the entry is Address and packets which are addressed to that address are
    sent to the interface specified in device*/


   int ioctlcall;
   struct rtentry request;
   struct sockaddr_in *sock, *sockmask;

   bzero ((char *) &request, sizeof (struct rtentry));

   sock = (struct sockaddr_in *) &request.rt_dst;
   sock->sin_family = AF_INET;
   sock->sin_addr.s_addr = target;
   
   sockmask = (struct sockaddr_in *) &request.rt_genmask;
   sockmask->sin_family = AF_INET;
   sockmask->sin_addr.s_addr = Mask;

   ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

   request.rt_flags = RTF_UP | RTF_HOST;
   request.rt_dev = device;
   if (ioctl(ioctlcall, SIOCADDRT, (char *) &request) < 0)
	perror ("SIOCADDRT failed");   
   close (ioctlcall);

}

/********************************************************************/

void routeEntry (int type, unsigned long target, unsigned long dest,
		 unsigned long Mask)
{

/*This function is used to add (type = 1)  and remove (type = 2)  routes in 
  the MN*/

int ioctlsid;
struct rtentry rtent;
struct arpreq ar;
struct sockaddr_in *sock, *socket2, *sockMask;

bzero ((char *)&rtent, sizeof (struct rtentry));

sock = (struct sockaddr_in *)&rtent.rt_dst;
sock->sin_family = AF_INET;
sock->sin_addr.s_addr = target;

socket2 = (struct sockaddr_in *)&rtent.rt_gateway;
socket2->sin_family = AF_INET;
socket2->sin_addr.s_addr = dest;

if (type == 1)
{
  sockMask = (struct sockaddr_in *)&rtent.rt_genmask;
  sockMask->sin_family = AF_INET;
  sockMask->sin_addr.s_addr = Mask;
}
ioctlsid = socket (AF_INET, SOCK_DGRAM, 0);

switch (type)
{
 case 1:
   
     rtent.rt_flags = RTF_UP |  RTF_GATEWAY;

     if (ioctl(ioctlsid, SIOCADDRT, (char *)&rtent) == -1)
        perror ("SIOCADDRT failed");
     break;
  

 case 2:
  
     if (ioctl(ioctlsid, SIOCDELRT, (char *) &rtent) < 0 && errno!= EFAULT)
        perror ("SIOCDELRT failed"); 
     break; 

}
close (ioctlsid);
}

/*********************************************************************/

void delRouteAndARP()
{
   FILE *router;
   char string[300],device[20], ipaddress[20];
   unsigned long dest, gateway, mask;
   unsigned int a1, a2, a3, a4;

/*This function is used to remove all the entries in /proc/net/route and all
the entries in the ARP cache (/proc/net/arp)*/

   router = fopen("/proc/net/route","r");
        
   fgets (string, 255, router);
   while (fgets(string, 255, router) != NULL)
     {

        if (sscanf(string, "%s %lx %lx %u %u %u %u %lx ", device, &dest, &gateway,
                &a1, &a2, &a3, &a4, &mask) != 8) continue;
 
        if (strcmp(device,"lo"))
            routeEntry (2, dest, gateway, mask);
     }
   fclose (router);

   router = fopen ("/proc/net/arp","r");
   fgets(string, 255, router);
   while (fgets(string, 255, router) != NULL)
     {
        if (sscanf (string, "%s ",ipaddress) != 1) continue;
        DEBUG (2) printf ("Address to delete from ARP cache: %s\n",ipaddress);
        if ((dest = inet_addr(ipaddress)) != -1) 
	  {
	       delARPentry(dest,0);
	       delARPentry(dest,1);	
          }
	else DEBUG(2) printf ("Address %s not deleted from the arp cache\n",ipaddress);
     }
   fclose (router);
}

/********************************************************************************/

void delARPentry (unsigned long ipaddress, int permanent)
{

/*Given one IP address, this function deletes that IP address from the ARP cache*/


   struct arpreq request;
   struct sockaddr_in *sock;
   int ioctlsid;
   
   ioctlsid = socket (AF_INET,  SOCK_DGRAM, 0);

   bzero ((char *)&request, sizeof(struct arpreq));
   sock = (struct sockaddr_in *) &request.arp_pa;
   sock->sin_family = AF_INET;
   sock->sin_addr.s_addr = ipaddress;
   /*if (permanent) request.arp_flags = ATF_COM; 
   */if (ioctl(ioctlsid, SIOCDARP, (char *) &request) == -1)
      perror ("It cannot delete an entry in ARP cache");
   else DEBUG (2)
            printf ("Entry removed ...\n");
   close (ioctlsid);
}

#include <net/if_arp.h>
#include <linux/if_ether.h>

/********************************************************************/

void devARPrequest (int type, char *device)
{

  /*This procedure is used to force the network interface (specified in 
    device) to answer to ARP requests (type = 2) or the other way round 
    (type = 1)
    */

   struct sockaddr_in *sock;
   struct ifreq request;
   int ioctlcall;

   bzero ((char *)&request, sizeof (struct ifreq));
   ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
   
   bcopy (device, request.ifr_name, 6);
   if (ioctl(ioctlcall, SIOCGIFFLAGS, (char *)&request) < 0)
     DEBUG (1) perror ("Error getting device flags:");

   switch (type)
    {
     case 1: /*No arp flag*/
       request.ifr_flags |= IFF_NOARP;
       if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) < 0)
         DEBUG (1) perror ("Error setting device flags: ");
       break;
     case 2: /*Setting an ARP flag */
       request.ifr_flags &= ~IFF_NOARP;
       if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) < 0)
         DEBUG (1) perror ("Error setting device flags: ");
       break;
    }
   close (ioctlcall);
}


/********************************************************************/

void createARP (unsigned long Address, char *hwAddress)
{

/*This function creates an ARP entry (permanent!)*/

  int ioctlcall;
  struct sockaddr_in *sock;
  struct arpreq request;

  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);
 
  bzero ((char *)&request, sizeof (struct arpreq));
  sock = (struct sockaddr_in *) &request.arp_pa;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address;

  sock = (struct sockaddr_in *)&request.arp_ha;
  sock->sin_family = ARPHRD_ETHER;
  bcopy (hwAddress, request.arp_ha.sa_data, 6);
  request.arp_flags = (ATF_PERM | ATF_COM);
  if (ioctl(ioctlcall, SIOCSARP, (char *)&request) < 0)
      perror ("ARP has not been done");
  close (ioctlcall);
}

/********************************************************************/

void addIFtunl (int tunnel, unsigned long Address)
{

  struct sockaddr_in *sock;
  struct ifreq request;
  int ioctlcall,count;
  char device[10];
  
  bzero ((char *)&request, sizeof (struct ifreq));
  ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

  count = sprintf (device, "tunl%c", tunnel+'0');
  device[count] = '\0';
  
  bcopy (device, request.ifr_name, 6);
  sock = (struct sockaddr_in *)&request.ifr_addr;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = Address; /*Means FA*/
  if (ioctl(ioctlcall, SIOCSIFADDR, (char *)&request) == -1)
    perror ("Error creating a tunnel");

  sock = (struct sockaddr_in *)&request.ifr_dstaddr;
  sock->sin_family = AF_INET;
  sock->sin_addr.s_addr = mynode.HA.Addr;
  
  if (ioctl(ioctlcall, SIOCSIFDSTADDR, (char *)&request) == -1)
    perror ("Error creating a tunnel (2)");

  request.ifr_flags = (IFF_NOARP | IFF_UP | IFF_POINTOPOINT);
  if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) == -1)
    perror ("Error creating the tunnel (3)");
 
  close (ioctlcall);

}

/********************************************************************/

void delIFtunl (int tunnel)
{

   char device[10];
   int ioctlcall;
   struct ifreq request;

   bzero ((char *)&request, sizeof (struct ifreq));
   ioctlcall = socket (AF_INET, SOCK_DGRAM, 0);

   device[sprintf (device, "tunl%c",tunnel + '0')] = '\0';
   bcopy (device, request.ifr_name, 6);
   
   if (ioctl(ioctlcall, SIOCSIFADDR, (char *)&request) <0)
     perror ("tunnel can not be deleted");
   
   request.ifr_flags &= ~IFF_UP;
   if (ioctl(ioctlcall, SIOCSIFFLAGS, (char *)&request) == -1)
     perror ("tunnel can not be deleted (2)");
   close (ioctlcall);

}

/********************************************************************/

